#!/usr/bin/env python3
import os
import docker
import json
import logging as logger
import platformdirs as plat
import shutil
import uuid
from ia.gaius.agent_client import AgentClient
from ia.scripts.spawn_agent import stopit, start_agent


class AgentInfo:
    def __init__(self, user_id: str, agent_id: str, agent_name: str, genome_name: str = None, genome_file: str = None, agent_config = None, location: str = 'local', LOG_LEVEL=logger.INFO, api_key="ABCD-1234"):
        """Wrapper class for Agent information. Used to start/stop agents, and save config information to a central location

        Args:
            user_id (str): User id attached to docker contaienrs spawned using AgentInfo
            agent_id (str): Agent id attached to docker containers spawned using AgentInfo
            agent_name (str): Descriptive, user-friendly name to refer to agent by
            genome_name (str, optional): If genome already in centralized location, refer to it by its name
            genome_file (str, optional): To copy outside genome into genome store, provide filepath here
            agent_config (_type_, optional): Stores AgentConfig details once agent is spawned.
            location (str, optional): In future will support spawning at several locations (local, private cloud, DigitalOcean, etc). Defaults to 'local'.
            LOG_LEVEL (logging.LEVEL, optional): Used to control verbosity of messages in agent spawn script. Defaults to logging.INFO.
            api_key (str, optional): Used to set API KEY of spawned agent. Defaults to "ABCD-1234"
        """
        logger.basicConfig(level=logger.INFO)
        self.data_dir = plat.user_data_dir(appname="IA_SDK_AgentManager", appauthor="IntelligentArtifacts")
        self.genome_dir = f'{self.data_dir}/genomes'
        self.agents_dir = f'{self.data_dir}/agents'
        
        if genome_name:
           self.genome_name =  genome_name
        else:
            self.genome_name = os.path.basename(genome_file)
            if os.path.exists(f'{self.genome_dir}/{self.genome_name}'):
                logger.warn(f'copying genome to genome dir would overwrite current genome {os.path.basename(genome_file)}, using genome already present')
            else:
                shutil.copyfile(src=genome_file, dst=f'{self.genome_dir}/{self.genome_name}')

        self.genome_file =  f'{self.genome_dir}/{self.genome_name}'
        self.user_id = user_id
        self.agent_id = agent_id
        self.location = location
        self.agent_name = agent_name
        self.agent_config = agent_config
        self.api_key = api_key
        
    def check_running(self):
        """Check if we can access the agent

        Returns:
            bool: True for success, False for failure
        """
        if self.location == 'local':

            agent = AgentClient(bottle_info=self.agent_config)
            try:
                agent.connect()
            except Exception:
                return False

            return True

        else:
            raise Exception(f'unknown location {self.location}')
    
    def delete_config_file(self):
        """Kill agent and delete config information from agents directory
        """
        self.kill_agent()
        if os.path.exists(f'{self.agents_dir}/{self.agent_name}'):
            os.remove(f'{self.agents_dir}/{self.agent_name}')
        return f"deleted {self.agents_dir}/{self.agent_name}"
        
    def save_config_file(self):
        """Store config information for agent in agents directory, based on agent name
        """
        with open(f'{self.agents_dir}/{self.agent_name}', 'w') as f:
            json.dump(obj=self.toJSON(), fp=f)

        print(f'saved to {self.agents_dir}/{self.agent_name}')
        return '{self.agents_dir}/{self.agent_name}'
    
    def kill_agent(self):
        """Attempt to kill a running agent
        """
        stopit(self.genome_file, user_id=self.user_id, agent_id=self.agent_id)
    
    def spawn(self, connect_jia=True):
        """Start a new agent using information provided in constructor

        Args:
            api_key (str, optional): API Key for agent. Defaults to "ABCD-1234".
        """
        with open(self.genome_file, 'r') as f:
            genome = json.load(f)
        
        json_obj = {'agent_id': self.agent_id,
                    'user_id': self.user_id,
                    'kill': False,
                    'network': None,
                    'api_key': self.api_key,
                    'genome': genome}

        self.agent_config, ports = start_agent(json_obj=json_obj)
        print(f'updating config file with new agent config: {self.agent_config}')
        self.save_config_file()
        
        if connect_jia:
            self.connect_jia()

        return json.loads(json.dumps(self.agent_config))

    def get_docker_networks(self):
        """Get names of networks this agent is a part of (uses gaius-api-{user_id}-{agent_id})

        Raises:
            Exception: More that one container found with same name

        Returns:
            _type_: list of networks
        """
        docker_client = docker.from_env()
        gapi_container = docker_client.containers.list(filters={'name':f'gaius-api-{self.user_id}-{self.agent_id}'})
        if len(gapi_container) == 1:
            gapi_container = gapi_container[0]
            return list(gapi_container.attrs['NetworkSettings']['Networks'].keys())
        elif len(gapi_container) > 1:
            self.delete_config_file()
            raise Exception('too many docker containers with same GAIuS API name')
        else:
            raise Exception('gaius-api container not found, start agent first')
        
    
    def connect_jia(self):
        """Connect agent to jia (add jia to the agent's network)
        """
        docker_client = docker.from_env()
        jias = docker_client.containers.list(filters={'ancestor':'registry.digitalocean.com/intelligent-artifacts/jia:0.1.15'})
        nets = self.get_docker_networks()
        nets = docker_client.networks.list(names=nets)
        if len(jias) == 0:
            print("No JIA notebooks found! Skipping...")
        elif len(jias) == 1:
            print(f'connecting single jia')
            jia = docker_client.containers.get(jias[0].short_id) #.attrs['Config']['Name']
            for net in nets:
                net.reload()
                if jia.id in [cont.id for cont in net.containers]:
                    print(f'jia already connected to network {net.name}')
                    continue
                net.connect(jia)
                print(f"Connecting JIA id: {jia.name} on {net.name}")
        elif len(jias) > 1:
            def connect_jia_func():
                print("More than one JIA found. Please select which to connect:")
                for c, i in enumerate(jias):
                    x = int(input(f"    {i}: {c} {c.name} connect? Y/N - "))
                    if x.lowercase() == 'y':
                        for net in nets:
                            net.reload()
                            if c.id in [cont.id for cont in net.containers]:
                                print(f'jia already connected to network {net.name}')
                                continue
                            net.connect(c)
                            print(f"Connecting JIA id: {c.name} on {net.name}")
                    else:
                        print("Invalid entry. Try again...")
    
    def get_agent_client(self):
        """Retreive AgentClient object from started agent

        Returns:
            ia.gaius.agent_client.AgentClient
        """
        if self.agent_config == None:
            raise Exception('agent_config information empty, must start agent first')
        return AgentClient(bottle_info=self.agent_config)
    
    def toJSON(self):
        """Dump config information to a JSON object (dict)

        Returns:
            dict: config information
        """
        return {'genome_name': self.genome_name,
                'genome_file': self.genome_file,
                'agent_id': self.agent_id,
                'user_id': self.user_id,
                'agent_name': self.agent_name,
                'location': self.location,
                'agent_config': self.agent_config,
                'api_key': self.api_key
                }
    
    @classmethod
    def fromFile(cls, filepath: str):
        """Construct AgentInfo object from file

        Args:
            filepath (str): path to file to load

        Returns:
            AgentInfo: AgentInfo object created from the file
        """
        with open(filepath, 'r') as f:
            json_obj = json.load(f)
        
        return cls.fromJSON(json_obj=json_obj)
    
    @classmethod
    def fromJSON(cls, json_obj):
        """Construct AgentInfo from JSON object

        Args:
            json_obj (dict): Contains fields necessary for constructing AgentInfo

        Returns:
            AgentInfo: AgentInfo object created from JSON object
        """
        return cls(**json_obj)


class AgentManager:
    def __init__(self):
        """Initialize AgentManager object (setup dirs, etc.)
        
        This object can be used to spawn Agents using the start_agent() method.
        """
        self.data_dir = plat.user_data_dir(appname="IA_SDK_AgentManager", appauthor="IntelligentArtifacts")
        self.genome_dir = f'{self.data_dir}/genomes'
        self.agents_dir = f'{self.data_dir}/agents'
        self.config_file = f'{self.data_dir}/config.json'
        
        if not os.path.exists(self.data_dir):
            os.mkdir(self.data_dir)
        print(f'user data location: {self.data_dir}')
        
        if not os.path.exists(self.genome_dir):
            os.mkdir(self.genome_dir)
            
        if not os.path.exists(self.agents_dir):
            os.mkdir(self.agents_dir)
        
        self.update_current_agents()
        
    def start_agent(self, genome_file=None, genome_name=None, user_id: str = 'jia', agent_id: str = '1', agent_name = None, connect_jia=True, api_key='ABCD-1234'):
        """Spawn an agent and save its information in the localized agents dir
        
        .. note:: 

            If you wish to use a genome that has not been used by the AgentManager before,
            pass it in the genome_file field. This will retrieve the genome from the specified path,
            and save it in the localized genome dir for future use.


        Args:
            genome_file (str, optional): Path to outside genome file. Defaults to None.
            genome_name (str, optional): Name of genome in the localized genome dir. Defaults to None.
            user_id (str, optional): user id for naming docker containers. Defaults to 'jia'.
            agent_id (str, optional): agent id for naming docker containers. Defaults to '1'.
            agent_name (_type_, optional): descriptive name "alias" for agent. Defaults to None.
            connect_jia (bool, optional): Whether to attempt to connect agent to Jia docker container. Defaults to True.
            api_key (str, optional): Control the api key used to access agent. Defaults to 'ABCD-1234'.

        Raises:
            Exception: If there is a conflicting agent with user_id and agent_id already present/alive

        Returns:
            AgentInfo: AgentInfo object corresponding to started agent. can be used to retrieve
            AgentClient object using AgentInfo.get_agent_client()
        """
        if agent_name is None:
            agent_name = f'agent-{uuid.uuid4().hex[:8]}'
        self.update_current_agents()
        for name, agent in self.current_agents.items():
            if agent.agent_id == agent_id and agent.user_id == user_id:
                raise Exception(f'Agent({name}) with user_id {user_id} and agent_id {agent_id} already exists, please choose other ids')
        
        if agent_name is None:
            f'agent-{uuid.uuid4().hex[:8]}'
        agent = AgentInfo(agent_id=agent_id, user_id=user_id, genome_file=genome_file, genome_name=genome_name, agent_name=agent_name, api_key=api_key)
        agent.spawn(connect_jia=True)
        self.update_current_agents()
        print(f'started agent {agent.agent_name}')
        return agent
    
    def get_all_agent_status(self):
        """Get the status of all agents in current_agents dict

        Returns:
            dict: { agent_name : alive? }
        """
        agent_status = {}
        for agent_name, agent in self.current_agents.items():
            agent_status[agent_name] = agent.check_running()

        return agent_status
        
    def delete_agent(self, agent_name):
        """Stop/Delete agent & remove config file from agents_dir

        Args:
            agent_name (str): name of agent to stop

        Returns:
            str: string depicting deletion status
        """
        result = self.current_agents[agent_name].delete_config_file()
        del self.current_agents[agent_name]
        self.update_current_agents()
        return result
    
    def list_genomes(self):
        """List the genomes present in the localized genome store

        Returns:
            list: list of genome files 
        """
        return os.listdir(self.genome_dir)
    
    def delete_genome(self, genome_name):
        """Delete genome from centralized genome_dir

        Args:
            genome_name (str): genome name to delete
        """
        genome_path = os.path.join(self.genome_dir, genome_name)
        print(f'will remove genome {genome_path}')
        os.remove(genome_path)
        return
    
    def get_genome(self, genome_name):
        """Get genome by name, return as dict

        Args:
            genome_name (str): filename for genome in genome_dir

        Raises:
            Exception: When genome not found

        Returns:
            dict: contents of genome file
        """
        genome_path = os.path.join(self.genome_dir, genome_name)
        
        if genome_name not in self.list_genomes():
            raise Exception(f'{genome_name} not in genome dir: {self.list_genomes()}')
        with open(genome_path, 'r') as f:
            return json.load(f)
        
    
    def update_current_agents(self):
        """Update the current_agents_dict
        """

        # retrieve all agent files
        self.current_agents = os.listdir(self.agents_dir)
        
        # get full paths of agent files
        self.current_agents = [os.path.join(self.agents_dir, agent) for agent in self.current_agents]
        
        # instantiate AgentInfo classes
        self.current_agents = [AgentInfo.fromFile(agent) for agent in self.current_agents]
        
        # make current agents into a dict by agent_name
        self.current_agents = {agent.agent_name : agent for agent in self.current_agents}
        
    def remediate_dead_agents(self):
        """Remove all agents that cannot be pinged.
        """
        
        agent_status = self.get_all_agent_status()
        
        killed_agents = []
        for agent_name, alive in agent_status.items():
            if not alive:
                print(f'{agent_name} is not alive, cleaning up')
                self.delete_agent(agent_name)
                killed_agents.append(agent_name)
        
        print(f'{killed_agents=}')

    def kill_all_agents(self):
        """Kill all agents known to AgentManager
        """
        for agent_name in list(self.current_agents.keys()):
            self.delete_agent(agent_name=agent_name)


import math
import functools
import json
import operator
from copy import deepcopy
from collections import Counter, defaultdict
from uuid import uuid4
from itertools import chain
from traceback import print_exc
import matplotlib.pyplot as plt
import pandas as pd

# Data
from sklearn.metrics import accuracy_score
import numpy as np

# Gaius Agent
from ia.gaius.agent_client import AgentClient
from ia.gaius.pvt.mongo_interface import MongoData, MongoResults
from ia.gaius.data_ops import Data
from ia.gaius.pvt.pvt_utils import retrieve_emotive_val, emotives_polarity_metrics_builder, \
    emotives_value_metrics_builder, make_modeled_emotives_, init_emotive_on_node, classification_metrics_builder, plot_confusion_matrix
from ia.gaius.prediction_models import hive_model_classification, average_emotives, prediction_ensemble_model_classification

class PVTAbortError(Exception):
    "Raised when PVT is aborted by Celery. Used to exit cleanly from nested test/train functions"

class PVTMessage():
    """Wrapper for PVT socket messages to be sent during training and testing"""
    def __init__(self, status: str,
                 current_record: int,
                 total_record_count: int,
                 metrics: dict,
                 cur_test_num: int,
                 total_test_num: int,
                 test_id: str = None,
                 user_id: str = '',
                 test_type: str = 'default'):
        self.status = status
        self.current_record = current_record
        self.total_record_count = total_record_count
        self.metrics = metrics
        self.test_id = test_id
        self.user_id = user_id
        self.cur_test_num = cur_test_num
        self.total_test_num = total_test_num
        self.test_type = test_type

    def toJSON(self):
        return {'status': self.status,
                'current_record': self.current_record,
                'total_record_count': self.total_record_count,
                'metrics': self.metrics,
                'test_id': self.test_id,
                'user_id': self.user_id,
                'cur_test_num': self.cur_test_num,
                'total_test_num': self.total_test_num,
                'test_type': self.test_type
                }


class PerformanceValidationTest():
    """
    Performance Validation Test (PVT) - Splits a GDF folder into training and testing sets.
    Based on the test type certain visualizations will be produced.

    Test types:

    - Classification
    - Emotive Value
    - Emotives Polarity
    """
    def __init__(self,
                 agent: AgentClient,
                 ingress_nodes: list,
                 query_nodes: list,
                 num_of_tests: int,
                 pct_of_ds: float,
                 pct_res_4_train: float,
                 test_type: str,
                 dataset_location: str = 'filepath',
                 results_filepath=None,
                 ds_filepath: str = None,
                 test_prediction_strategy="continuous",
                 clear_all_memory_before_training: bool = True,
                 turn_prediction_off_during_training: bool = False,
                 shuffle: bool = False,
                 sio=None,
                 task=None,
                 user_id: str = None,
                 mongo_db=None,
                 dataset_info: dict = None,
                 test_id=None,
                 test_configuration: dict = {},
                 socket_channel: str = 'pvt_status',
                 QUIET: bool = False):
        """Initialize the PVT object with all required parameters for execution

        Args:
            agent (AgentClient): GAIuS Agent to use for trainings
            ingress_nodes (list): Ingress nodes for the GAIuS Agent (see :func:`ia.gaius.agent_client.AgentClient.set_query_nodes`)
            query_nodes (list): Query nodes for the GAIuS Agent (see :func:`ia.gaius.agent_client.AgentClient.set_query_nodes`)
            num_of_tests (int): Number of test iterations to complete
            pct_of_ds (float): Percent of the dataset to use for PVT (overall)
            pct_res_4_train (float): Percent of the dataset to be reserved for training
            test_type (str): classification, emotives_value, or emotives_polarity
            dataset_location (str): Location of dataset to utilise, "mongodb", or "filepath"
            results_filepath (_type_): Where to store PVT results
            ds_filepath (str): Path to the directory containing training GDFs
            test_prediction_strategy (str, optional): _description_. Defaults to "continuous".
            clear_all_memory_before_training (bool, optional): Whether the GAIuS agent's memory should be cleared before each training. Defaults to True.
            turn_prediction_off_during_training (bool, optional): Whether predictions should be disabled during training to reduce computational load. Defaults to False.
            shuffle (bool, optional): Whether dataset should be shuffled before each test iteration. Defaults to False.
            sio (_type_, optional): SocketIO object to emit information on. Defaults to None.
            task (_type_, optional): Celery details to emit information about. Defaults to None.
            user_id (str, optional): user_id to emit information to on SocketIO. Defaults to ''.
            mongo_db (pymongo.MongoClient, optional): MongoDB where dataset should be retrieved from
            dataset_info (dict, optional): information about how to retrieve dataset, used for MongoDB query. If dataset_location is mongodb, this must have the user_id, dataset_id, results_collection, logs_collection, and data_files_collection_name keys
            test_id (str, optional): unique identifier to be sent with messages about this test. Also used for storing to mongodb
            test_configuration (dict, optional): dictionary storing additional metadata about test configuration, to be saved in mongodb with test results
            socket_channel (str, optional): SocketIO channel to broadcast results on. Defaults to 'pvt_status'
            QUIET (bool, optional): flag used to disable log output during PVT. Defaults to False
        """

        self.agent                               = agent
        self.ingress_nodes                       = ingress_nodes
        self.query_nodes                         = query_nodes
        self.num_of_tests                        = num_of_tests
        self.dataset_location                    = dataset_location
        self.ds_filepath                         = ds_filepath
        self.results_filepath                    = results_filepath
        self.pct_of_ds                           = pct_of_ds
        self.pct_res_4_train                     = pct_res_4_train
        self.shuffle                             = shuffle
        self.test_type                           = test_type
        self.clear_all_memory_before_training    = clear_all_memory_before_training
        self.turn_prediction_off_during_training = turn_prediction_off_during_training
        self.test_prediction_strategy            = test_prediction_strategy

        self.emotives_set                        = None
        self.labels_set                          = None
        self.predictions                         = None
        self.actuals                             = None
        self.emotives_metrics_data_structures    = None
        self.class_metrics_data_structures       = None
        self.metrics_dataframe                   = None
        self.pvt_results                         = None
        self.sio                                 = sio
        self.task                                = task
        self.user_id                             = user_id
        self.mongo_db                            = mongo_db
        self.dataset_info                        = dataset_info
        self.test_id                             = test_id
        self.testing_log                         = []
        self.mongo_results                       = None
        self.test_configuration                  = test_configuration
        self.labels_counter                      = Counter()
        self.socket_channel                      = socket_channel
        self.QUIET                               = QUIET

        self.agent_genes = {}

        all_nodes = [node['name'] for node in self.agent.all_nodes]
        self.agent_genes = self.agent.get_all_genes(nodes=all_nodes)
        self.test_configuration['initial_agent_genes'] = self.agent_genes
        self.test_configuration['genome'] = self.agent.genome.topology

        if dataset_location == 'mongodb':
            self.dataset = MongoData(mongo_dataset_details=self.dataset_info, data_files_collection_name=self.dataset_info['data_files_collection_name'], mongo_db=mongo_db)
            self.mongo_results = MongoResults(mongo_db=self.mongo_db, result_collection_name=self.dataset_info['results_collection'],
                                              log_collection_name=self.dataset_info['logs_collection'], test_id=self.test_id, user_id=self.user_id,
                                              dataset_id=self.dataset_info['dataset_id'], test_configuration=self.test_configuration)
        elif dataset_location == 'filepath':
            self.dataset = Data(data_directories=[self.ds_filepath])
        elif dataset_location == 'prepared':
            self.dataset = self.ds_filepath
        elif dataset_location == 'prepared_obj':
            self.dataset = self.ds_filepath

        else:
            raise Exception(f'unknown value for dataset location: {dataset_location}')

        # Show Agent status by Default
        self.agent.show_status()

        # Assign Ingress and Query Nodes
        self.agent.set_ingress_nodes(nodes=self.ingress_nodes)
        self.agent.set_query_nodes(nodes=self.query_nodes)

        # Setting summarize single to False by default in order to handle multiply nodes topologies
        self.agent.set_summarize_for_single_node(False)
        
        if not self.QUIET: # pragma: no cover
            print(f"num_of_tests      = {self.num_of_tests}")
            print(f"ds_filepath       = {self.ds_filepath}")
            print(f"pct_of_ds         = {self.pct_of_ds}")
            print(f"pct_res_4_train   = {self.pct_res_4_train}")
            print(f"summarize_for_single_node status   = {self.agent.summarize_for_single_node}")

    def prepare_datasets(self):
        if self.dataset_location == 'prepared':
            if not self.QUIET: # pragma: no cover
                print(f"Length of Training Set = {len(self.dataset.train_sequences)}\n")
                print(f"Length of Testing Set  = {len(self.dataset.test_sequences)}\n")
            return

        elif self.dataset_location == 'prepared_obj':
            if not self.QUIET: # pragma: no cover
                print(f"Length of Training Set = {len(self.dataset.train_sequences)}\n")
                print(f"Length of Testing Set  = {len(self.dataset.test_sequences)}\n")
            return
        else:
            self.dataset.prep(
                percent_of_dataset_chosen=self.pct_of_ds,
                percent_reserved_for_training=self.pct_res_4_train,
                shuffle=self.shuffle
            )
            if not self.QUIET: # pragma: no cover
                print(f"Length of Training Set = {len(self.dataset.train_sequences)}\n")
                print(f"Length of Testing Set  = {len(self.dataset.test_sequences)}\n")
        return

    def run_classification_pvt(self):
        for test_num in range(0, self.num_of_tests):
            self.test_num = test_num
            if not self.QUIET: # pragma: no cover
                print(f'Conducting Test # {test_num}')
                print('\n---------------------\n')

            self.prepare_datasets()

            if self.sio:
                self.sio.emit(self.socket_channel,
                              PVTMessage(status='training', current_record=0, total_record_count=len(self.dataset.train_sequences),
                                         metrics={}, cur_test_num=self.test_num, total_test_num=self.num_of_tests,
                                         test_id=self.test_id, user_id=self.user_id, test_type=self.test_type).toJSON(), to=self.user_id)
            try:

                self.train_agent()

                if self.pct_res_4_train == 100:
                    print(f'Complete!')
                    continue

                self.test_agent()

                for k, labels in self.labels_set.items():
                    self.labels_set[k] = set([label.rsplit('|', maxsplit=1)[-1] for label in labels])
                if not self.QUIET: # pragma: no cover
                    print('Getting Classification Metrics...')
                self.get_classification_metrics()
                if not self.QUIET: # pragma: no cover
                    print('Saving results to pvt_results...')
                self.pvt_results.append(deepcopy(self.class_metrics_data_structures))
                self.pvt_results[test_num] = deepcopy(self.update_test_results_w_hive_classification_metrics(self.pvt_results[test_num]))

            except Exception as e:
                if not self.QUIET: # pragma: no cover
                    print('error during training/testing phase of test, remediating database for failed test, then raising error')
                if self.mongo_results:
                    if not self.QUIET: # pragma: no cover
                        print('about to remediate database')
                    self.mongo_results.deleteResults()
                    if not self.QUIET: # pragma: no cover
                        print('remediated database')

                if not self.QUIET: # pragma: no cover
                    print(f'raising error {str(e)}')
                raise e

            try:
                if self.dataset_location != 'mongodb':
                    if not self.QUIET: # pragma: no cover
                        print('Plotting Results...')
                    plot_confusion_matrix(test_num=test_num, class_metrics_data_structures=self.class_metrics_data_structures)
            except Exception as e:
                if not self.QUIET: # pragma: no cover
                    print(f'error plotting results from classification pvt: {e}')
                pass

            response_dict = {'classification_counter': self.labels_counter,
                             'pvt_results': self.pvt_results}
            result_msg = PVTMessage(status='finished', current_record=0, total_record_count=0,
                                    metrics=response_dict, cur_test_num=self.test_num, total_test_num=self.num_of_tests,
                                    test_id=self.test_id, user_id=self.user_id, test_type=self.test_type).toJSON()
            if self.sio:
                self.sio.emit(self.socket_channel,
                                result_msg, to=self.user_id)

        if self.mongo_results:
            self.mongo_results.saveResults(result_msg)
        return

    def run_emotive_value_pvt(self):
        self.pvt_results = []
        for test_num in range(0, self.num_of_tests):

            if not self.QUIET: # pragma: no cover
                print(f'Conducting Test # {test_num}')
                print('\n---------------------\n')

            self.prepare_datasets()
            if self.sio:
                self.sio.emit(self.socket_channel,
                              PVTMessage(status='training', current_record=0, total_record_count=len(self.dataset.train_sequences),
                                         metrics={}, cur_test_num=self.test_num, total_test_num=self.num_of_tests,
                                         test_id=self.test_id, user_id=self.user_id, test_type=self.test_type).toJSON(), to=self.user_id)

            self.train_agent()

            if self.pct_res_4_train == 100:
                print(f'Complete!')
                return

            self.test_agent()
            
            if not self.QUIET: # pragma: no cover
                print('Getting Emotives Value Metrics...')
            self.get_emotives_value_metrics()
            if not self.QUIET: # pragma: no cover
                print('Saving results to pvt_results...')
            self.pvt_results.append(deepcopy(self.emotives_metrics_data_structures))
            self.pvt_results[test_num] = deepcopy(self.update_test_results_w_hive_emotives_value_metrics(self.pvt_results[test_num]))
            if not self.QUIET: # pragma: no cover
                print('Plotting Results...')

            # don't try to plot emotive values if we're working to save in a mongo database
            # (its probably running without a jupyter GUI)
            if self.mongo_db is None:
                self.plot_emotives_value_charts(test_num=test_num)

        # send out finished socket message
        response_dict = {'pvt_results': self.pvt_results}

        final_msg = PVTMessage(status='finished', current_record=0, total_record_count=0,
                                     metrics=response_dict, cur_test_num=self.num_of_tests, total_test_num=self.num_of_tests,
                                     test_id=self.test_id, user_id=self.user_id, test_type=self.test_type).toJSON()
        if self.sio:
            self.sio.emit(self.socket_channel,
                          final_msg, to=self.user_id)
        if self.mongo_results:
            self.mongo_results.saveResults(final_msg)
        return

    def run_emotive_polarity_pvt(self):
        self.pvt_results = []
        for test_num in range(0, self.num_of_tests):
            if not self.QUIET: # pragma: no cover
                print(f'Conducting Test # {test_num}')
                print('\n---------------------\n')
            self.prepare_datasets()

            if self.sio:
                self.sio.emit(self.socket_channel,
                              PVTMessage(status='training', current_record=0, total_record_count=len(self.dataset.train_sequences),
                                         metrics={}, cur_test_num=self.test_num, total_test_num=self.num_of_tests,
                                         test_id=self.test_id, user_id=self.user_id, test_type=self.test_type).toJSON(), to=self.user_id)

            if not self.QUIET: # pragma: no cover
                print("Training Agent...")
            self.train_agent()

            if self.pct_res_4_train == 100:
                print(f'Complete!')
                return

            if not self.QUIET: # pragma: no cover
                print("Testing Agent...")
            self.test_agent()

            if not self.QUIET: # pragma: no cover
                print('Getting Emotives Polarity Metrics...')
            if not self.QUIET: # pragma: no cover
                print('Saving results to pvt_results...')
            self.pvt_results.append(deepcopy(self.get_emotives_polarity_metrics()))

        # send out finished socket message
        response_dict = {'pvt_results': self.pvt_results}
        final_msg = PVTMessage(status='finished', current_record=0, total_record_count=0,
                                     metrics=response_dict, cur_test_num=self.num_of_tests, total_test_num=self.num_of_tests,
                                     test_id=self.test_id, user_id=self.user_id, test_type=self.test_type).toJSON()
        if self.sio:
            self.sio.emit(self.socket_channel,
                          final_msg, to=self.user_id)
        if self.mongo_results:
            self.mongo_results.saveResults(final_msg)
        return

    def conduct_pvt(self):
        """
        Function called to execute the PVT session. Determines test to run based on 'test_type' attribute

        Results from PVT is stored in the 'pvt_results' attribute

        .. note::

            A complete example is shown in the :func:`__init__` function above. Please see that documentation for further information about how to conduct a PVT test

        """
        try:
            self.test_num = 0
            self.pvt_results = []
            self.testing_log = []

            # Validate Test Type
            if self.test_type == 'classification':
                if not self.QUIET: # pragma: no cover
                    print("Conducting Classification PVT...\n")
                self.run_classification_pvt()

            elif self.test_type == 'emotives_value':
                if not self.QUIET: # pragma: no cover
                    print("Conducting Emotives Value PVT...\n")
                self.run_emotive_value_pvt()

            elif self.test_type == 'emotives_polarity':
                if not self.QUIET: # pragma: no cover
                    print("Conducting Emotives Polarity PVT...\n")
                self.run_emotive_polarity_pvt()

            else:
                raise Exception(
                    """
                    Please choose one of the test type:
                    - classification
                    - emotives_value
                    - emotives_polarity

                    ex.
                    --> pvt.test_type='emotives_value'
                    then, retry
                    --> pvt.conduct_pvt()
                    """
                )
        # except PVTAbortError as e:
        #     print(f'Aborting PVT: {str(e)}')
        #     return
        except Exception as e:
            if not self.QUIET: # pragma: no cover
                print('\n--------------------------------------------------------------\n')
                print(f'failed to conduct PVT test, test_type={self.test_type}: {str(e)}')
                print_exc()
                print('\n--------------------------------------------------------------\n')
            raise e

    def train_agent(self):
        """
        Takes a training set of gdf files, and then trains an agent on those records.
        The user can turn prediction off if the topology doesn't have abstractions
        where prediction is needed to propagate data through the topology.
        """
        # Initialize
        if self.clear_all_memory_before_training is True:
            if not self.QUIET: # pragma: no cover
                print('Clearing memory of selected ingress nodes...')
            self.agent.clear_all_memory(nodes=self.ingress_nodes)

        if self.test_type == 'classification':
            # Start an Labels Tracker for each node
            if not self.QUIET: # pragma: no cover
                print('Initialize labels set...')
            self.labels_set = {}
            for node in self.ingress_nodes:
                self.labels_set[node] = set()
            if not self.QUIET: # pragma: no cover
                print(self.labels_set)
            self.labels_counter.clear()
            if not self.QUIET: # pragma: no cover
                print('Created labels set...')
        elif self.test_type == 'emotives_value' or self.test_type == 'emotives_polarity':
            # Start an Emotives Tracker for each node
            if not self.QUIET: # pragma: no cover
                print('Initialize emotives set...')
            self.emotives_set = {}
            for node in self.ingress_nodes:
                self.emotives_set[node] = set()
            if not self.QUIET: # pragma: no cover
                print(self.emotives_set)
                print('Created emotives set...')
        else:
            raise Exception(f"Invalid test type {self.test_type}")
        # Train Agent
        if self.turn_prediction_off_during_training is True:
            self.agent.stop_predicting(nodes=self.query_nodes)
        else:
            self.agent.start_predicting(nodes=self.query_nodes)
        if not self.QUIET: # pragma: no cover
            print('Preparing to train agent...')
        # for i, file_path in enumerate(log_progress(dataset.train_sequences)):

        train_seq_len = len(self.dataset.train_sequences)

        train_metrics = {}
        if self.test_type == 'classification':
            train_metrics = {'classification_counter': self.labels_counter}

        for j, _ in enumerate(self.dataset.train_sequences):

            training_msg = PVTMessage(status='training', current_record=j + 1, total_record_count=train_seq_len,
                                      metrics=train_metrics, cur_test_num=self.test_num, total_test_num=self.num_of_tests,
                                      test_id=self.test_id, user_id=self.user_id, test_type=self.test_type)

            self.store_train_record(test_num=self.test_num, record=training_msg)

            if j % 10 == 0:
                if self.task:
                    if self.task.is_aborted():
                        self.abort_test_remediation(current_record=j, record_count=train_seq_len)
                        return


            if j % 100 == 0:
                if not self.QUIET: # pragma: no cover
                    print(f"train - {j}")
            if self.dataset_location == 'filepath':
                with open(self.dataset.train_sequences[j], "r") as sequence_file:
                    sequence = sequence_file.readlines()
                    sequence = [json.loads(d) for d in sequence]
            elif self.dataset_location == 'prepared':
                with open(self.dataset.train_sequences[j], "r") as sequence_file:
                    sequence = sequence_file.readlines()
                    sequence = [json.loads(d) for d in sequence]
            elif self.dataset_location == 'prepared_obj':
                sequence = self.dataset.train_sequences[j]
            elif self.dataset_location == 'mongodb':
                sequence = self.dataset.getSequence(self.dataset.train_sequences[j])
            else:
                raise Exception(f"dataset location {self.dataset_location} is unknown")

            for event in sequence:
                self.agent.observe(data=event, nodes=self.ingress_nodes)
                if self.test_type == 'emotives_value' or self.test_type == 'emotives_polarity':
                    for node in self.ingress_nodes:
                        percept_emotives = list(self.agent.get_percept_data()[node]['emotives'].keys())
                        # print(f'updating emotive set for {node} with {percept_emotives}')
                        self.emotives_set[node].update(percept_emotives)
            if self.test_type == 'classification':
                for node in self.ingress_nodes:
                    self.labels_set[node].update(sequence[-1]['strings'])
                self.labels_counter.update([label.rsplit('|', maxsplit=1)[-1] for label in sequence[-1]['strings']])
            self.agent.learn(nodes=self.ingress_nodes)
        if not self.QUIET: # pragma: no cover
            print('Finished training agent!')

    def test_agent(self):
        """
        Test agent on dataset test sequences provided in self.dataset.test_sequences
        """
        # Start Testing
        self.agent.start_predicting(nodes=self.query_nodes)
        self.predictions = []
        self.actuals     = []

        self.testing_log.append([])
        test_step_info = {}
        test_seq_len = len(self.dataset.test_sequences)
        if test_seq_len == 0:
            if not self.QUIET: # pragma: no cover
                print('length of testing sequences is 0... returning\n')
            return
        for k, _ in enumerate(self.dataset.test_sequences):

            if k % 10 == 0:
                if self.task:
                    if self.task.is_aborted():
                        self.abort_test_remediation(current_record=k, record_count=test_seq_len)
                        return

            if k % 100 == 0:
                if not self.QUIET: # pragma: no cover
                    print(f"test - {k}")
            if self.dataset_location == 'filepath':
                with open(self.dataset.test_sequences[k], "r") as sequence_file:
                    sequence = sequence_file.readlines()
                    sequence = [json.loads(d) for d in sequence]
            elif self.dataset_location == 'prepared':
                with open(self.dataset.test_sequences[k], "r") as sequence_file:
                    sequence = sequence_file.readlines()
                    sequence = [json.loads(d) for d in sequence]
            elif self.dataset_location == 'prepared_obj':
                sequence = self.dataset.test_sequences[k]
            elif self.dataset_location == 'mongodb':
                sequence = self.dataset.getSequence(self.dataset.test_sequences[k])
            else:
                raise Exception(f"dataset location {self.dataset_location} is unknown")

            self.agent.clear_wm(nodes=self.ingress_nodes)
            if self.test_type == 'classification':
                # observe up to last event, which has the answer
                if len(sequence) > 2:
                    self.agent.stop_predicting(nodes=self.ingress_nodes)
                    for event in sequence[:-2]:
                        self.agent.observe(data=event, nodes=self.ingress_nodes)
                    self.agent.start_predicting(nodes=self.ingress_nodes)
                    self.agent.observe(data=sequence[-2], nodes=self.ingress_nodes)
                else:
                    for event in sequence[:-1]:
                        self.agent.observe(data=event, nodes=self.ingress_nodes)
                # get and store predictions after observing events
                self.predictions.append(self.agent.get_predictions(nodes=self.query_nodes))
                # store answers in a separate list for evaluation
                classifications_split = [label.rsplit('|', maxsplit=1)[-1] for label in sequence[-1]['strings']]
                self.actuals.append(deepcopy(classifications_split))
                for node in self.ingress_nodes:
                    self.labels_set[node].update(sequence[-1]['strings'])
                self.labels_counter.update([label.rsplit('|', maxsplit=1)[-1] for label in sequence[-1]['strings']])
                # get predicted classification on the fly, so we can save to mongo individually
                pred_dict = {node: self.predictions[k][node] for node in self.query_nodes}
                for key in pred_dict:
                    if pred_dict[key] is None:
                        pred_dict[key] = 'UNKNOWN'
                    else:
                        pred_dict[key] = prediction_ensemble_model_classification(pred_dict[key]).most_common()[0][0]
                test_step_info.update({'idx': k, 'predicted': pred_dict, 'actual': self.actuals[k],
                                       'classification_counter': self.labels_counter})
                test_step_info = self.compute_incidental_probabilities(test_step_info=test_step_info)

                # observe answer
                self.agent.observe(sequence[-1], nodes=self.ingress_nodes)

            elif self.test_type == 'emotives_value' or self.test_type == 'emotives_polarity':
                for event in sequence:
                    self.agent.observe(data=event, nodes=self.ingress_nodes)
                    for node in self.ingress_nodes:
                        self.emotives_set[node].update(list(self.agent.get_percept_data()[node]['emotives'].keys()))
                # get and store predictions after observing events
                self.predictions.append(self.agent.get_predictions(nodes=self.query_nodes))
                # store answers in a separate list for evaluation
                self.actuals.append(self.sum_sequence_emotives(sequence))

                pred_dict = {node: self.predictions[k][node] for node in self.query_nodes}
                for key in pred_dict:
                    pred_dict[key] = make_modeled_emotives_(pred_dict[key])

                test_step_info.update({'idx': k, 'predicted': pred_dict, 'actual': self.actuals[-1]})
                test_step_info = self.compute_incidental_probabilities(test_step_info=test_step_info)

            else:
                raise Exception(f"Invalid test type {self.test_type}")

            # prepare test step message
            test_step_msg = PVTMessage(status='testing', current_record=k + 1, total_record_count=test_seq_len,
                                       metrics=test_step_info, cur_test_num=self.test_num, total_test_num=self.num_of_tests,
                                       test_id=self.test_id, user_id=self.user_id, test_type=self.test_type)

            self.store_train_record(test_num=self.test_num, record=test_step_msg)

            # learn answer (optional continous learning)
            if self.test_prediction_strategy == "continuous":
                self.agent.learn(nodes=self.ingress_nodes)
            elif self.test_prediction_strategy == "noncontinuous":
                continue
            else:
                raise Exception(
                    """
                    Not a valid test prediction strategy. Please choose either 'continuous',
                    which means to learn the test sequence/answer after the agent has tried to make a prediction on that test sequence,
                    or, 'noncontinuous', which means to not learn the test sequence.
                    """
                )

    def sum_sequence_emotives(self, sequence):
        """
        Sums all emotive values
        """
        emotives_seq = [event['emotives'] for event in sequence if event['emotives']]
        return dict(functools.reduce(operator.add, map(Counter, emotives_seq)))

    def get_classification_metrics(self):
        """
        Builds classification data structures for each node
        """
        self.class_metrics_data_structures = {}
        for node, labels in self.labels_set.items():
            self.class_metrics_data_structures[node] = classification_metrics_builder(lst_of_labels=labels)
            # Let's see how well the agent scored
            overall_preds = []
            answers       = []
            for p in range(0, len(self.predictions)):
                overall_pred = prediction_ensemble_model_classification(self.predictions[p][node])
                if overall_pred is None:
                    # if the agent doesn't have enough information to make a prediction it wouldn't give one
                    overall_preds.append('UNKNOWN')
                else:
                    overall_preds.append(overall_pred.most_common()[0][0])
                answers.append(self.actuals[p][0])
            try:
                print(f'{answers=}')
                print(f'{overall_preds=}')
                accuracy = round(accuracy_score(answers, overall_preds), 2) * 100
            except ZeroDivisionError:
                accuracy = 0.0
            prec_predictions       = [p for p, a in zip(overall_preds, answers) if p != 'UNKNOWN']
            prec_answers           = [a for p, a in zip(overall_preds, answers) if p != 'UNKNOWN']
            try:
                if len(prec_predictions) == 0:
                    precision = 0.0
                else:
                    precision = round(accuracy_score(prec_answers, prec_predictions), 2) * 100
            except ZeroDivisionError:
                precision = 0.0
            total_amount_of_questions = len(answers)
            updated_pred_length       = len([p for p in overall_preds if p != 'UNKNOWN'])
            try:
                resp_pc = np.round(updated_pred_length / total_amount_of_questions, 2) * 100
            except ZeroDivisionError:
                resp_pc = 0.0

            if math.isnan(precision):
                precision = 0.0

            self.class_metrics_data_structures[node]['predictions']          = overall_preds
            self.class_metrics_data_structures[node]['actuals']              = answers
            self.class_metrics_data_structures[node]['metrics']['resp_pc']   = resp_pc
            self.class_metrics_data_structures[node]['metrics']['accuracy']  = accuracy
            self.class_metrics_data_structures[node]['metrics']['precision'] = precision

    def compute_incidental_probabilities(self, test_step_info: dict):
        """Keep track of how well each node is doing during the testing phase. To be used for live visualizations

        Args:
            test_step_info (dict, required): Dictionary containing information about the current predicted, actual answers, and other related metrics (e.g. precision, unknowns, residuals, response rate, etc.)

        Returns:
            dict: updated test_step_info with the current running accuracy
        """
        idx = test_step_info['idx']

        if self.test_type == 'classification':

            # compute hive prediction for time idx
            hive_pred = hive_model_classification(ensembles=self.predictions[idx])
            hive_pred = hive_pred.most_common()[0][0]
            if hive_pred is None:
                hive_pred = 'UNKNOWN'
            test_step_info['predicted']['hive'] = hive_pred

            if 'running_accuracy' not in test_step_info:
                test_step_info['true_positive'] = dict()
                test_step_info['false_positive'] = dict()
                test_step_info['unknown_percentage'] = dict()

                test_step_info['running_accuracy'] = dict()
                for k in test_step_info['predicted'].keys():
                    if test_step_info['predicted'][k] in test_step_info['actual']:
                        test_step_info['running_accuracy'][k] = 1.0
                        test_step_info['true_positive'][k] = 1
                        test_step_info['false_positive'][k] = 0

                    else:
                        test_step_info['running_accuracy'][k] = 0.0
                        test_step_info['true_positive'][k] = 0
                        test_step_info['false_positive'][k] = 1

                test_step_info['response_percentage'] = dict()
                test_step_info['response_counts'] = dict()
                test_step_info['running_precisions'] = dict()
                for k in test_step_info['predicted'].keys():
                    if test_step_info['predicted'][k] != 'UNKNOWN':
                        test_step_info['response_percentage'][k] = 1.0
                        test_step_info['response_counts'][k] = 1
                        test_step_info['running_precisions'][k] = test_step_info['running_accuracy'][k]
                    else:
                        test_step_info['response_percentage'][k] = 0.0
                        test_step_info['running_precisions'][k] = 1.0
                        test_step_info['response_counts'][k] = 0

                    test_step_info['unknown_percentage'][k] = 1 - test_step_info['response_percentage'][k]

            else:
                for k in test_step_info['predicted'].keys():
                    if test_step_info['predicted'][k] != 'UNKNOWN':
                        test_step_info['response_counts'][k] += 1

                    if test_step_info['predicted'][k] in test_step_info['actual']:
                        test_step_info['true_positive'][k] += 1
                    else:
                        test_step_info['false_positive'][k] += 1

                    try:
                        test_step_info['running_precisions'][k] = test_step_info['true_positive'][k] / test_step_info['response_counts'][k]
                    except ZeroDivisionError:
                        test_step_info['running_precisions'][k] = 0.0
                        pass

                    test_step_info['running_accuracy'][k] = test_step_info['true_positive'][k] / (idx + 1)
                    test_step_info['response_percentage'][k] = test_step_info['response_counts'][k] / (idx + 1)
                    test_step_info['unknown_percentage'][k] = 1 - test_step_info['response_percentage'][k]

        elif self.test_type == 'emotives_polarity':

            # compute hive prediction for time idx
            test_step_info['predicted']['hive'] = average_emotives(list(test_step_info['predicted'].values()))

            # Initialize metrics on first observation
            if 'running_accuracy' not in test_step_info:
                test_step_info['true_positive']       = defaultdict(dict)
                test_step_info['false_positive']      = defaultdict(dict)
                test_step_info['true_negative']       = defaultdict(dict)
                test_step_info['false_negative']      = defaultdict(dict)
                test_step_info['unknown_percentage']  = defaultdict(dict)
                test_step_info['response_percentage'] = defaultdict(dict)
                test_step_info['response_counts']     = defaultdict(dict)
                test_step_info['running_accuracy']    = defaultdict(dict)
                test_step_info['running_precisions']  = defaultdict(dict)

                for k in test_step_info['predicted'].keys():
                    for emotive in test_step_info['actual'].keys():
                        if emotive in test_step_info['predicted'][k].keys():
                            pred_sign = np.sign(test_step_info['predicted'][k][emotive])
                            actual_sign = np.sign(test_step_info['actual'][emotive])

                            # If actual sign is zero for a polarity test, something is wrong in the data
                            if actual_sign == 0:
                                raise Exception(f'Zero value found in polarity test at idx {idx}')

                            # Check if the predicted sign is zero (unknown) to incremenet response count and percentage
                            if bool(pred_sign):
                                test_step_info['response_counts'][k][emotive] = 1
                                test_step_info['response_percentage'][k][emotive] = 1.0
                                test_step_info['unknown_percentage'][k][emotive] = 0.0
                            else:
                                test_step_info['response_counts'][k][emotive] = 0
                                test_step_info['response_percentage'][k][emotive] = 0.0
                                test_step_info['unknown_percentage'][k][emotive] = 1.0

                            # True positive (correct)
                            if actual_sign > 0 and pred_sign > 0:
                                test_step_info['true_positive'][k][emotive] = 1
                                test_step_info['true_negative'][k][emotive] = 0
                                test_step_info['false_positive'][k][emotive] = 0
                                test_step_info['false_negative'][k][emotive] = 0

                                test_step_info['running_accuracy'][k][emotive] = 1.0

                            # True Negative (correct)
                            elif actual_sign < 0 and not pred_sign > 0:
                                test_step_info['true_positive'][k][emotive] = 0
                                test_step_info['true_negative'][k][emotive] = 1
                                test_step_info['false_positive'][k][emotive] = 0
                                test_step_info['false_negative'][k][emotive] = 0

                                test_step_info['running_accuracy'][k][emotive] = 1.0

                            # False positive (incorrect)
                            elif actual_sign < 0 and not pred_sign < 0:
                                test_step_info['true_positive'][k][emotive] = 0
                                test_step_info['true_negative'][k][emotive] = 0
                                test_step_info['false_positive'][k][emotive] = 1
                                test_step_info['false_negative'][k][emotive] = 0

                                test_step_info['running_accuracy'][k][emotive] = 0.0

                            # False negative (incorrect)
                            elif actual_sign > 0 and not pred_sign > 0:
                                test_step_info['true_positive'][k][emotive] = 0
                                test_step_info['true_negative'][k][emotive] = 0
                                test_step_info['false_positive'][k][emotive] = 0
                                test_step_info['false_negative'][k][emotive] = 1

                                test_step_info['running_accuracy'][k][emotive] = 0.0

                        # Unknown answer
                        else:
                            test_step_info['response_counts'][k][emotive] = 0
                            test_step_info['response_percentage'][k][emotive] = 0.0
                            test_step_info['unknown_percentage'][k][emotive] = 1.0

                # Initialize precision
                for k in test_step_info['predicted'].keys():
                    for emotive in test_step_info['actual'].keys():
                        if emotive in test_step_info['predicted'][k].keys():
                            if bool(test_step_info['predicted'][k][emotive]):
                                test_step_info['running_precisions'][k][emotive] = test_step_info['running_accuracy'][k]
                            else:
                                test_step_info['running_precisions'][k][emotive] = 1.0

            # Not the first observation, so update values
            else:
                for k in test_step_info['predicted'].keys():
                    for emotive in test_step_info['actual'].keys():

                        # catch new emotives, not yet seen on node {k}
                        if emotive not in test_step_info['true_positive'][k].keys():
                            init_emotive_on_node(emotive=emotive, test_step_info=test_step_info, node=k)
                        if emotive in test_step_info['predicted'][k].keys():
                            pred_sign = np.sign(test_step_info['predicted'][k][emotive])
                            actual_sign = np.sign(test_step_info['actual'][emotive])

                            # else:
                                # print(f'predicted val: {test_step_info["predicted"][k][emotive]}, actual: {test_step_info["actual"][emotive]}')
                            # Polarity error at for zero actual value
                            if actual_sign == 0:
                                raise Exception(f'Zero value found in polarity test at idx {idx}')

                            # If predicted value non-zero
                            if bool(pred_sign):
                                test_step_info['response_counts'][k][emotive] += 1

                            # True positive (correct)
                            if actual_sign > 0 and pred_sign > 0:
                                test_step_info['true_positive'][k][emotive] += 1

                            # True Negative (correct)
                            elif actual_sign < 0 and pred_sign < 0:
                                test_step_info['true_negative'][k][emotive] += 1

                            # False positive (incorrect)
                            elif actual_sign < 0 and not pred_sign < 0:
                                test_step_info['false_positive'][k][emotive] += 1

                            # False negative (incorrect)
                            elif actual_sign > 0 and not pred_sign > 0:
                                test_step_info['false_negative'][k][emotive] += 1

                            # Calculate precision value
                            try:
                                test_step_info['running_precisions'][k][emotive] = (test_step_info['true_positive'][k][emotive] + test_step_info['true_negative'][k][emotive]) / test_step_info['response_counts'][k][emotive]
                            except ZeroDivisionError:
                                test_step_info['running_precisions'][k][emotive] = 0.0

                            test_step_info['running_accuracy'][k][emotive] = (test_step_info['true_positive'][k][emotive] + test_step_info['true_negative'][k][emotive]) / (idx + 1)

                        # Update response percentage and unknown percentage
                        test_step_info['response_percentage'][k][emotive] = test_step_info['response_counts'][k][emotive] / (idx + 1)
                        test_step_info['unknown_percentage'][k][emotive] = 1 - test_step_info['response_percentage'][k][emotive]

        elif self.test_type == 'emotives_value':
            test_step_info['predicted']['hive'] = average_emotives(list(test_step_info['predicted'].values()))
            # Initialize metrics on first observation
            if 'running_accuracy' not in test_step_info:
                test_step_info['unknown_percentage']  = defaultdict(dict)
                test_step_info['response_percentage'] = defaultdict(dict)
                test_step_info['response_counts']     = defaultdict(dict)
                test_step_info['residuals']           = defaultdict(dict)
                test_step_info['abs_residuals']       = defaultdict(dict)
                test_step_info['squared_residuals']   = defaultdict(dict)

                for k in test_step_info['predicted'].keys():
                    for emotive in test_step_info['actual'].keys():
                        if emotive in test_step_info['predicted'][k].keys():
                            test_step_info['residuals'][k][emotive] = test_step_info['actual'][emotive] - test_step_info['predicted'][k][emotive]
                            test_step_info['abs_residuals'][k][emotive] = abs(test_step_info['actual'][emotive] - test_step_info['predicted'][k][emotive])
                            test_step_info['squared_residuals'][k][emotive] = math.pow(test_step_info['actual'][emotive] - test_step_info['predicted'][k][emotive], 2)
                            test_step_info['response_counts'][k][emotive] = 1
                            test_step_info['response_percentage'][k][emotive] = 1.0
                            test_step_info['unknown_percentage'][k][emotive] = 0.0

                        # Unknown answer
                        else:
                            test_step_info['residuals'][k] = defaultdict(float)
                            test_step_info['response_counts'][k][emotive] = 0
                            test_step_info['response_percentage'][k][emotive] = 0.0
                            test_step_info['unknown_percentage'][k][emotive] = 1.0

            # Not the first observation, so update values
            else:
                for emotive in test_step_info['actual'].keys():
                    if emotive in test_step_info['predicted'][k].keys():

                        test_step_info['residuals'][k][emotive] = test_step_info['actual'][emotive] - test_step_info['predicted'][k][emotive]
                        test_step_info['response_counts'][k][emotive] += 1

                    # Update response percentage and unknown percentage
                    test_step_info['response_percentage'][k][emotive] = test_step_info['response_counts'][k][emotive] / (idx + 1)
                    test_step_info['unknown_percentage'][k][emotive] = 1 - test_step_info['response_percentage'][k][emotive]

        else:
            raise Exception('Unknown test type')

        return test_step_info

    def get_emotives_value_metrics(self):
        """
        Builds emotives value data structures for each node
        """
        # Build an emotives Metric Data Structure
        self.emotives_metrics_data_structures = {}
        for node, emotive_set in self.emotives_set.items():
            self.emotives_metrics_data_structures[node] = emotives_value_metrics_builder(lst_of_emotives=list(emotive_set))

        # Populate Emotives Metrics
        for i, (prediction_ensemble, actual) in enumerate(zip(self.predictions, self.actuals)):
            for node_name, node_pred_ensemble in prediction_ensemble.items():
                if node_pred_ensemble:
                    modeled_emotives = make_modeled_emotives_(ensemble=node_pred_ensemble)  # get overall prediction from a single node
                    for emotive_name_from_model, pred_value in modeled_emotives.items():
                        if emotive_name_from_model in list(self.emotives_metrics_data_structures[node_name].keys()):
                            self.emotives_metrics_data_structures[node_name][emotive_name_from_model]['predictions'].append(pred_value)
                            self.emotives_metrics_data_structures[node_name][emotive_name_from_model]['actuals'].append(retrieve_emotive_val(emotive_name_from_model, actual))
                    left_overs = set(self.emotives_metrics_data_structures) - set(list(modeled_emotives.keys()))
                    if left_overs:
                        for emotive_name, metric_data in self.emotives_metrics_data_structures[node_name].items():
                            if emotive_name in left_overs:
                                self.emotives_metrics_data_structures[node_name][emotive_name]['predictions'].append(np.nan)
                                self.emotives_metrics_data_structures[node_name][emotive_name]['actuals'].append(retrieve_emotive_val(emotive_name, actual))
                else:
                    for emotive_name, metric_data in self.emotives_metrics_data_structures[node_name].items():
                        self.emotives_metrics_data_structures[node_name][emotive_name]['predictions'].append(np.nan)
                        self.emotives_metrics_data_structures[node_name][emotive_name]['actuals'].append(retrieve_emotive_val(emotive_name, actual))
                # Create Metrics
                for node_name, node_emotive_metrics in self.emotives_metrics_data_structures.items():
                    # calculate response rate percentage
                    for emotive_name, data in node_emotive_metrics.items():
                        total_amount_of_questions            = len(data['actuals'])
                        updated_pred_length                  = len([p for p in data['predictions'] if p is not np.nan])
                        try:
                            resp_pc = np.round(updated_pred_length / total_amount_of_questions, 2) * 100
                        except ZeroDivisionError:
                            resp_pc = 0.0
                        self.emotives_metrics_data_structures[node_name][emotive_name]['metrics']['resp_pc'] = resp_pc
                    # calculate rmse
                    for emotive_name, data in node_emotive_metrics.items():
                        error_lst = [p - a for p, a in zip(data['predictions'], data['actuals']) if p is not np.nan]
                        if error_lst:
                            rmse = math.sqrt(np.square(error_lst).mean())
                            self.emotives_metrics_data_structures[node_name][emotive_name]['metrics']['rmse'] = rmse
                    # calculate smape_precision
                    for emotive_name, data in node_emotive_metrics.items():
                        smape_prec_predictions       = [p for p, a in zip(data['predictions'], data['actuals']) if p is not np.nan]
                        smape_prec_actuals           = [a for p, a in zip(data['predictions'], data['actuals']) if p is not np.nan]
                        smape_prec_predictions_array = np.array(smape_prec_predictions)
                        smape_prec_actuals_array     = np.array(smape_prec_actuals)

                        forecast_diff = np.abs(smape_prec_predictions_array - smape_prec_actuals_array)
                        forecast_sum  = (np.abs(smape_prec_actuals_array) + np.abs(smape_prec_predictions_array)) / 2
                        smape_elements = forecast_diff / forecast_sum

                        try:
                            smape = (1 - (np.nansum(smape_elements) / len(smape_prec_actuals_array))) * 100
                        except ZeroDivisionError:
                            smape = None
                        self.emotives_metrics_data_structures[node_name][emotive_name]['metrics']['smape_prec'] = smape

    def get_emotives_polarity_metrics(self):
        """
        Builds emotives polarity data structures for each node
        """
        template_dict = {'true_positive': {},
                         'false_positive': {},
                         'true_negative': {},
                         'false_negative': {},
                         'unknown_percentage': {},
                         'response_percentage':{},
                         'response_counts': {},
                         'running_accuracy': {},
                         'running_precisions': {}
                        }
        if len(self.testing_log[self.test_num]) == 0:
            return {}
        # lets flip the dictionary so that it is organized per node instead of per metric
        raw_test_results = deepcopy(self.testing_log[self.test_num][-1]['metrics'])

        flattened_emotive_set = set(chain(*[list(item) for item in self.emotives_set.values()]))
        hive_emotives = flattened_emotive_set
        emo_res2 = {k: {i: deepcopy(template_dict) for i in v } for k, v in self.emotives_set.items()}
        emo_res2['hive'] = {emo: deepcopy(template_dict) for emo in hive_emotives}
        emotive_results = {k: deepcopy(template_dict) for k in flattened_emotive_set}
        # print(f'before: {emotive_results=}')
        for k, v in raw_test_results.items():
            if k in ['idx', 'predicted', 'actual']:
                continue

            for node, info in v.items():
                for emotive, emotive_data in info.items():
                    emotive_results[emotive][k][node] = deepcopy(emotive_data)
                    if emotive in emo_res2[node]:
                        emo_res2[node][emotive][k] = deepcopy(emotive_data)


        return emo_res2

        # Build an emotives Metric Data Structure
        self.emotives_metrics_data_structures = {}
        for node, emotive_set in self.emotives_set.items():
            self.emotives_metrics_data_structures[node] = emotives_polarity_metrics_builder(lst_of_emotives=list(emotive_set))
        # Populate Emotives Metrics
        for i, (prediction_ensemble, actual) in enumerate(zip(self.predictions, self.actuals)):
            for node_name, node_pred_ensemble in prediction_ensemble.items():
                if node_pred_ensemble:
                    modeled_emotives = make_modeled_emotives_(ensemble=node_pred_ensemble)  # get overall prediction from a single node
                    for emotive_name_from_model, pred_value in modeled_emotives.items():
                        if emotive_name_from_model in list(self.emotives_metrics_data_structures[node_name].keys()):
                            self.emotives_metrics_data_structures[node_name][emotive_name_from_model]['predictions'].append(pred_value)
                            self.emotives_metrics_data_structures[node_name][emotive_name_from_model]['actuals'].append(retrieve_emotive_val(emotive_name_from_model, actual))
                    left_overs = set(self.emotives_metrics_data_structures) - set(list(modeled_emotives.keys()))
                    if left_overs:
                        for emotive_name, metric_data in self.emotives_metrics_data_structures[node_name].items():
                            if emotive_name in left_overs:
                                self.emotives_metrics_data_structures[node_name][emotive_name]['predictions'].append(np.nan)
                                self.emotives_metrics_data_structures[node_name][emotive_name]['actuals'].append(retrieve_emotive_val(emotive_name, actual))
                else:
                    for emotive_name, metric_data in self.emotives_metrics_data_structures[node_name].items():
                        self.emotives_metrics_data_structures[node_name][emotive_name]['predictions'].append(np.nan)
                        self.emotives_metrics_data_structures[node_name][emotive_name]['actuals'].append(retrieve_emotive_val(emotive_name, actual))
                # Create Metrics
                for node_name, node_emotive_metrics in self.emotives_metrics_data_structures.items():
                    # calculate response rate percentage
                    for emotive_name, data in node_emotive_metrics.items():
                        total_amount_of_questions            = len(data['actuals'])
                        updated_pred_length                  = len([p for p in data['predictions'] if p is not np.nan])
                        try:
                            resp_pc = np.round(updated_pred_length / total_amount_of_questions, 2) * 100
                        except ZeroDivisionError:
                            resp_pc = 0.0
                        self.emotives_metrics_data_structures[node_name][emotive_name]['metrics']['resp_pc'] = resp_pc
                    # calculate accuracy
                    for emotive_name, data in node_emotive_metrics.items():
                        polarity_accuracy_loading_dock = []
                        polarity_precision_loading_dock = []
                        for p, a in zip(data['predictions'], data['actuals']):
                            if p is np.nan:
                                polarity_accuracy_loading_dock.append("incorrect")
                            elif p is not np.nan:
                                if p * a > 0:
                                    polarity_accuracy_loading_dock.append("correct")
                                elif p * a < 0:
                                    polarity_accuracy_loading_dock.append("incorrect")
                            else:
                                raise Exception("Something is wrong with the data type...")
                        try:
                            accuracy = round(polarity_accuracy_loading_dock.count("correct") / len(polarity_accuracy_loading_dock), 2) * 100
                        except ZeroDivisionError():
                            accuracy = 0.0
                        # populate accuracy to data structure
                        self.emotives_metrics_data_structures[node_name][emotive_name]['metrics']['accuracy'] = accuracy
                        # calculate precision
                        for p, a in zip(data['predictions'], data['actuals']):
                            if p is np.nan:
                                continue
                            elif p is not np.nan:
                                if p * a > 0:
                                    polarity_precision_loading_dock.append("correct")
                                elif p * a < 0:
                                    polarity_precision_loading_dock.append("incorrect")
                            else:
                                raise Exception("Something is wrong with the data type...")
                        try:
                            precision = round(polarity_precision_loading_dock.count("correct") / len(polarity_precision_loading_dock),
                                              2) * 100
                        except ZeroDivisionError():
                            precision = 0.0
                        # populate precision to data structure
                        self.emotives_metrics_data_structures[node_name][emotive_name]['metrics']['precision'] = precision

    def update_test_results_w_hive_classification_metrics(self, pvt_test_result):
        """
        Update pvt test result metrics with hive classifications metrics
        """
        # add hive_metrics
        hive_metrics = {
            'predictions': [],
            'actuals': [],
            'labels': [],
            'metrics': {
                'resp_pc': None,
                'accuracy': None,
                'precision': None
            }
        }

        # get hive labels set
        hive_label_count = []
        for node_name, test_data in pvt_test_result.items():
            if node_name != 'hive':
                for label in test_data['labels']:
                    hive_label_count.append(label)

        hive_label_set_lst = list(set(hive_label_count))

        # add hive metrics dictionary to pvt results
        pvt_test_result['hive'] = hive_metrics

        pvt_test_result['hive']['labels'] = hive_label_set_lst

        # get predictions to get hive classification of all nodes
        for i in range(0, len(self.predictions)):
            pred = hive_model_classification(ensembles=self.predictions[i])
            if pred is not None:
                pred = pred.most_common()[0][0]
            else:
                pred = 'UNKNOWN'
            pvt_test_result['hive']['predictions'].append(deepcopy(pred))

        # get actuals of test
        for i in range(0, len(self.actuals)):
            pvt_test_result['hive']['actuals'].append(self.actuals[i][0])

        # get hive accuracy of test
        for node_name, test_data in pvt_test_result.items():
            if node_name == 'hive':
                try:
                    hive_accuracy = round(accuracy_score(test_data['actuals'], test_data["predictions"]), 2) * 100
                except ZeroDivisionError:
                    hive_accuracy = 0.0
        pvt_test_result['hive']['metrics']['accuracy'] = hive_accuracy

        # get hive precision of test
        for node_name, test_data in pvt_test_result.items():
            if node_name == 'hive':

                prec_predictions       = [p for p, _ in zip(test_data["predictions"], test_data['actuals']) if p != 'UNKNOWN']
                prec_answers           = [a for p, a in zip(test_data["predictions"], test_data['actuals']) if p != 'UNKNOWN']
                try:
                    hive_precision = round(accuracy_score(prec_answers, prec_predictions), 2) * 100
                except ZeroDivisionError:
                    hive_precision = 0.0
        pvt_test_result['hive']['metrics']['precision'] = hive_precision

        # get hive response rate percentage of test
        for node_name, test_data in pvt_test_result.items():
            if node_name == 'hive':
                total_amount_of_questions            = len(test_data['actuals'])
                updated_pred_length                  = len([p for p in test_data["predictions"] if p != 'UNKNOWN'])
                try:
                    hive_resp_pc = np.round(updated_pred_length / total_amount_of_questions, 2) * 100
                except ZeroDivisionError:
                    hive_resp_pc = 0.0
        pvt_test_result['hive']['metrics']['resp_pc'] = hive_resp_pc

        return pvt_test_result

    def update_test_results_w_hive_emotives_value_metrics(self, pvt_test_result):
        """
        Update pvt test result metrics with hive classifications metrics
        """
        all_nodes_emotives_set = set()
        for node, emotives in self.emotives_set.items():
            for emotive in emotives:
                all_nodes_emotives_set.add(emotive)
        hive_emotives_value_metrics_lst = []
        for emotive in all_nodes_emotives_set:
            hive_emotives_value_metrics_template = {
                f'{emotive}': {
                    'metrics': {
                        'resp_pc': [],
                        'rmse': [],
                        'smape_prec': []
                    },
                    # 'predictions' : {node: [] for node in self.emotives_set},
                    # 'actuals' : {node: [] for node in self.emotives_set}
                }
            }
            hive_emotives_value_metrics_lst.append(hive_emotives_value_metrics_template)

        for test_metric in self.pvt_results:
            for test_data in test_metric.values():
                for emotive, emotive_metric in test_data.items():
                    if emotive_metric['metrics']['resp_pc'] != 0.0 and emotive in all_nodes_emotives_set:
                        for hive_emotive_metric_dict in hive_emotives_value_metrics_lst:
                            if emotive in hive_emotive_metric_dict.keys():
                                hive_emotive_metric_dict[emotive]['metrics']['resp_pc'].append(emotive_metric['metrics']['resp_pc'])
                                hive_emotive_metric_dict[emotive]['metrics']['rmse'].append(emotive_metric['metrics']['rmse'])
                                hive_emotive_metric_dict[emotive]['metrics']['smape_prec'].append(emotive_metric['metrics']['smape_prec'])

        # average emotive metrics into one final result
        for emotive_metrics in hive_emotives_value_metrics_lst:
            for emotive_name, metric_data in emotive_metrics.items():
                for metric_name, metric_data_lst in metric_data['metrics'].items():
                    if len(metric_data_lst) == 0:
                        continue
                    emotive_metrics[emotive_name]['metrics'][metric_name] = sum(metric_data_lst) / len(metric_data_lst)

        # flatten list of dicts to single dict
        hive_emotive_overall_dict = {}
        for emo_dict in hive_emotives_value_metrics_lst:
            hive_emotive_overall_dict.update(**emo_dict)

        pvt_test_result['hive'] = hive_emotive_overall_dict
        return pvt_test_result

    def update_test_results_w_hive_emotives_polarity_metrics(self, pvt_test_result):
        """
        Update pvt test result metrics with hive emotives polarity metrics
        """
        all_nodes_emotives_set = set()
        for node, emotives in self.emotives_set.items():
            for emotive in emotives:
                all_nodes_emotives_set.add(emotive)
        hive_emotives_value_metrics_lst = []
        for emotive in all_nodes_emotives_set:
            hive_emotives_value_metrics_template = {
                f'{emotive}': {
                    'metrics': {
                        'resp_pc': [],
                        'accuracy': [],
                        'precision': []
                    }
                }
            }
            hive_emotives_value_metrics_lst.append(hive_emotives_value_metrics_template)
        for test_metric in self.pvt_results:
            for test_data in test_metric.values():
                for emotive, emotive_metric in test_data.items():
                    if emotive_metric['metrics']['resp_pc'] != 0.0 and emotive in all_nodes_emotives_set:
                        for hive_emotive_metric_dict in hive_emotives_value_metrics_lst:
                            if emotive in hive_emotive_metric_dict.keys():
                                hive_emotive_metric_dict[emotive]['metrics']['resp_pc'].append(emotive_metric['metrics']['resp_pc'])
                                hive_emotive_metric_dict[emotive]['metrics']['accuracy'].append(emotive_metric['metrics']['accuracy'])
                                hive_emotive_metric_dict[emotive]['metrics']['precision'].append(emotive_metric['metrics']['precision'])
        for emotive_metrics in hive_emotives_value_metrics_lst:
            for emotive_name, metric_data in emotive_metrics.items():
                for metric_name, metric_data_lst in metric_data['metrics'].items():
                    if len(metric_data_lst) == 0:
                        continue
                    emotive_metrics[emotive_name]['metrics'][metric_name] = sum(metric_data_lst) / len(metric_data_lst)

        hive_emotive_overall_dict = {}
        for emo_dict in hive_emotives_value_metrics_lst:
            hive_emotive_overall_dict.update(**emo_dict)

        if not self.QUIET: # pragma: no cover
            print(f'{hive_emotive_overall_dict=}')
        pvt_test_result['hive'] = hive_emotive_overall_dict
        return pvt_test_result

    def plot_emotives_value_charts(self, test_num):

        for node_name, node_emotive_metrics in self.emotives_metrics_data_structures.items():
            if not self.QUIET: # pragma: no cover
                print(f'-----------------Test#{test_num}-{node_name}-Plots-----------------')
            for emotive_name, data in sorted(node_emotive_metrics.items()):
                labels = 'precision', 'miss'
                if data['metrics']['smape_prec'] is None:
                    sizes = [0, 100]
                else:
                    sizes = [data['metrics']['smape_prec'], 100 - data['metrics']['smape_prec']]
                explode = (0, 0)
                _, ax1 = plt.subplots()
                ax1.title.set_text(f'{node_name} - {emotive_name}')
                ax1.pie(sizes, explode=explode, labels=labels, autopct='%1.1f%%',
                        shadow=True, startangle=90)
                ax1.axis('equal')  # Equal aspect ratio ensures that pie is drawn as a circle.
                colors = ['gray', 'skyblue']
                patches, texts = plt.pie(sizes, colors=colors, startangle=90)
                plt.legend(patches, labels, loc="best")
                plt.figtext(0, 0, f"{pd.Series(data['metrics']).round(1).to_string()}", ha="center", fontsize=18, bbox={"facecolor": "orange", "alpha": 0.5, "pad": 5})
                try:
                    plt.savefig(f"{self.results_filepath}/{test_num}_{node_name}_{emotive_name}.png", dpi=300, bbox_inches='tight')
                except Exception as e:
                    if not self.QUIET: # pragma: no cover
                        print(f"Not able to save figure in assigned results directory! Please add an appropriate directory: {str(e)}")
                    pass
                plt.show()
                if not self.QUIET: # pragma: no cover
                    print('---------------------')

    def abort_test_remediation(self, current_record, record_count):
        if not self.QUIET: # pragma: no cover
            print(f'about to abort {self.task.request.id =}, {self.test_id=}')
        if self.sio:
            if not self.QUIET: # pragma: no cover
                print('Sending abort message')
            abort_msg = PVTMessage(status='aborted', current_record=current_record + 1, total_record_count=record_count,
                                    metrics={}, cur_test_num=self.test_num, total_test_num=self.num_of_tests,
                                    test_id=self.test_id, user_id=self.user_id, test_type=self.test_type)
            self.sio.emit(self.socket_channel, abort_msg.toJSON(), to=self.user_id)
        if self.mongo_results:
            if not self.QUIET: # pragma: no cover
                print('cleaning up MongoDB')
            self.mongo_results.deleteResults()

        raise PVTAbortError(f"Aborting Test, at record {current_record} of {record_count}")

    def store_train_record(self, test_num, record: PVTMessage):
        
        if record.status == 'testing':
            self.testing_log[test_num].append(deepcopy(record.toJSON()))

        # insert into test_log in mongo, if using mongodb
        if self.mongo_results:
            self.mongo_results.addLogRecord(type=record.status, record=deepcopy(record.toJSON()))

        # emit socketIO message
        if self.sio:
            self.sio.emit(self.socket_channel, deepcopy(record.toJSON()), to=self.user_id)


"""Utilities for PVT computations"""
from sklearn.metrics import confusion_matrix, ConfusionMatrixDisplay
import matplotlib.pyplot as plt
from numpy import nan
from ia.gaius.prediction_models import average_emotives


def retrieve_emotive_val(emotive_name, actual):
    """Function to parse out emotive value from "actual" response.
    If emotive not present, return NaN

    Args:
        emotive_name (str): name of emotive to retrieve
        actual (dict): dictionary of actual emotive values from test record

    Returns:
        float: value of the specified emotive, or NaN if not present
    """
    if emotive_name in actual:
        return actual[emotive_name]
    else:
        return nan


def init_emotive_on_node(emotive: str, node: str, test_step_info: dict):
    """Helper function to initialize emotive information for live messages.
    Used if new emotive is encountered during testing
    (emotive only seen in specific records, not consistently across all)

    Args:
        emotive (str): emotive name
        node (str): node to initialize emotive on
        test_step_info (dict): dictionary of live information, which should
            be initialized with new emotive
    """
    test_step_info['true_positive'][node][emotive] = 0
    test_step_info['response_counts'][node][emotive] = 0
    test_step_info['true_positive'][node][emotive] = 0
    test_step_info['true_negative'][node][emotive] = 0
    test_step_info['false_positive'][node][emotive] = 0
    test_step_info['false_negative'][node][emotive] = 0
    return


def emotives_value_metrics_builder(lst_of_emotives: list) -> dict:
    """Create Metrics Data Structure for each emotive in testset

    Args:
        lst_of_emotives (list): emotives list to populate data structure

    Returns:
        dict: emotive metrics data structure
    """
    emotives_metrics_data_structure = {}
    for emotive in lst_of_emotives:
        emotives_metrics_data_structure[emotive] = {
            "predictions": [],
            "actuals": [],
            "metrics": {
                "resp_pc": None,  # response rate percentage
                "rmse": None,
                "smape_prec": None
            }
        }
    return emotives_metrics_data_structure


def emotives_polarity_metrics_builder(lst_of_emotives: list) -> dict:
    """
    Create Metrics Data Structure for each emotive in testset
    """
    emotives_metrics_data_structure = {}
    for emotive in lst_of_emotives:
        emotives_metrics_data_structure[emotive] = {
            "predictions": [],
            "actuals": [],
            "metrics": {
                "resp_pc": None,
                "accuracy": None,
                "precision": None
            }
        }
    return emotives_metrics_data_structure


def classification_metrics_builder(lst_of_labels: list) -> dict:
    """Create Metrics Data Structure for a classification problem
    where labels are tracked and used.

    Args:
        lst_of_labels (list): list of class labels

    Returns:
        dict: Classification data structure
    """
    classification_metrics_data_structure = {
        'predictions': [],
        'actuals': [],
        'labels': list(lst_of_labels) + ['UNKNOWN'],
        'metrics': {
            'resp_pc': None,
            'accuracy': None,
            'precision': None
        }
    }
    return classification_metrics_data_structure


def model_per_emotive_(ensemble: dict,
                       emotive: str,
                       potential_normalization_factor: float) -> float:
    """Using a Weighted Moving Average, though the 'moving' part
    refers to the prediction index.

    Args:
        ensemble (dict): prediction ensemble used to model
        emotive (str): emotive name to to model
        potential_normalization_factor (float): normalization factor

    Returns:
        float: final emotive modelled value
    """
    # using a weighted posterior_probability = potential/marginal_probability
    # FORMULA: pv + ( (Uprediction_2-pv)*(Wprediction_2) + (Uprediction_3-pv)*(Wprediction_3)... )/mp
    # A Weighted Moving Average puts more weight on recent data and less on past data. This is done by multiplying each bar’s price by a weighting factor. Because of its unique calculation, WMA will follow prices more closely than a corresponding Simple Moving Average.
    # https://www.fidelity.com/learning-center/trading-investing/technical-analysis/technical-indicator-guide/wma#:~:text=Description,a%20corresponding%20Simple%20Moving%20Average.
    _found = False
    while not _found:
        for i in range(0, len(ensemble)):
            if emotive in ensemble[i]['emotives'].keys():
                _found = True
                principal_value = ensemble[i]['emotives'][emotive]  # Let's use the "best" match (i.e. first showing of this emotive) as our starting point. Alternatively, we can use,say, the average of all values before adjusting.
                break
        if i == len(ensemble) and not _found:
            return 0
        if i == len(ensemble) and _found:
            return principal_value
    marginal_probability = sum([x["potential"] for x in ensemble])  # NOTE: marginal_probability = mp, this might the wrong calculation for this variable.
    weighted_moving_value = 0  # initialized top portion of summation
    for x in ensemble[i + 1:]:
        if emotive in x['emotives']:
            weighted_moving_value += (x['emotives'][emotive] - (principal_value)) * ((x["potential"] / potential_normalization_factor))
    weighted_moving_emotive_average = principal_value + (weighted_moving_value / marginal_probability)
    return weighted_moving_emotive_average


def make_modeled_emotives_(ensemble):
    '''The emotives in the ensemble are of type: 'emotives':[{'e1': 4, 'e2': 5}, {'e2': 6}, {'e1': 5 'e3': -4}]'''
    emotives_set = set()
    potential_normalization_factor = sum([p['potential'] for p in ensemble])

    filtered_ensemble = []
    for p in ensemble:
        new_record = p
        new_record['emotives'] = average_emotives([p['emotives']])  # AVERAGE
        filtered_ensemble.append(new_record)

    # filtered_ensemble = bucket_predictions(filtered_ensemble) # BUCKET

    for p in filtered_ensemble:
        emotives_set = emotives_set.union(p['emotives'].keys())
    return {emotive: model_per_emotive_(ensemble, emotive, potential_normalization_factor) for emotive in emotives_set}


def plot_confusion_matrix(test_num: int, class_metrics_data_structures: dict):
    """
    Takes a node classification test to create a confusion matrix.
    This version includes the i_dont_know or unknown label.
    """

    for node_name, class_metrics_data in class_metrics_data_structures.items():
        print(f'-------------Test#{test_num}-{node_name}-Plots-------------')
        sorted_labels = sorted(class_metrics_data['labels'])
        cm = confusion_matrix(class_metrics_data['actuals'],
                              class_metrics_data['predictions'],
                              labels=sorted_labels)
        disp = ConfusionMatrixDisplay(confusion_matrix=cm,
                                      display_labels=sorted_labels)
        disp.plot()
        if is_notebook():
            plt.show()
        else:
            cf_filename = f'confusion_matrix_{node_name}_test_{test_num}.png'
            print(f'attempting to save confusion_matrix to: {cf_filename}')
            plt.savefig(cf_filename)

def is_notebook() -> bool:
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True   # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False      # Probably standard Python interpreter


#!/usr/bin/env python3
import multiprocessing
from optparse import OptionParser
import os,sys
import json
import logging
import socket

import docker
from sty import fg, bg
import ia
from ia.gaius.genome_info import Genome

logging.basicConfig(format='%(asctime)s,%(msecs)03d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s',
datefmt='%Y-%m-%d:%H:%M:%S')
logger = logging.getLogger("spawn_agent")
logger.setLevel(level=logging.INFO)

ASSETS = {'network': {},
          'agent': None,
          'genome': None,
          'interfaces': {},
          'cp-containers': {},
          'mp-containers': {},
          'applications': {} ## These are connected containers as applications. Ex. JIA notebooks.
          }

def retrieve_genome(genome_file):
    logger.debug(f'{genome_file = }')
    
    # if provided filepath, attempt to open genome
    if os.path.exists(genome_file):
        with open(genome_file, 'r') as f:

            # read data into genome_json
            genome_json = f.read()
            _g = json.loads(genome_json)
            # convert json to python dictionary

    # otherwise try to load genome from string
    else:
        try:
            genome_json = genome_file
            _g = json.loads(genome_file)
        except json.decoder.JSONDecodeError:
            logger.critical(f'Failed to retreive genome, {genome_file=}')
            sys.exit(1)

    # remove description fields from genome
    _g['description'] = 'cleared description'
    for d in _g['elements']['nodes']:
        if 'description' in d['data'].keys():
            logger.debug(f"Removing: {d['data']['description']}")
            d['data']['description'] = ''

    return Genome(_g)

def parse_genome(genome):

    targets_map = {} ## values are target IDs for each key
    sources_map = {} ## values are source IDs of each key

    if 'edges' not in genome.topology['elements']:
        logger.warning("No connections.")
    else:
        for edge in genome.topology['elements']['edges']:
            d = edge['data']
            if d['source'] not in targets_map:
                if d['source'].startswith("action"):
                    continue
                targets_map[d['source']] = [d['target']]
            else:
                if d['source'].startswith("action"):
                    continue
                targets_map[d['source']].append(d['target'])
            if d['target'] not in sources_map:
                if d['target'].startswith("action"):
                    continue
                sources_map[d['target']] = [d['source']]
            else:
                if d['target'].startswith("action"):
                    continue
                sources_map[d['target']].append(d['source'])

    targets_map = eval( f"{targets_map}".replace(", 'P1_ID': ['action1_ID', 'action2_ID']", ""))

    as_inputs = {pid: [] for pid in genome.primitives.keys() }
    is_input = []

    for pid, p in genome.primitives.items():
        logger.debug(f"{pid} ({p['name']}):")
        logger.debug(f"    {p['sources']}")
        if p['sources'] == ['observables']:
            as_inputs[pid] = []
        else:
            for mid in p['manipulatives']:
                if mid.startswith("action"):
                    continue
                logger.debug(f"   {mid} {genome.manipulatives[mid]['name']} {genome.manipulatives[mid]['genes']['sources']['value']}")
                if genome.manipulatives[mid]['genes']['sources']['value'] == ['observables']:
                    as_inputs[pid].append(mid)
                    is_input.append(mid)

    logger.debug(f'{as_inputs=}')
    logger.debug(f'{is_input=}')


    logger.debug(f'{targets_map=}')
    ## for any manipulative that is an observables (i.e. in is_input list),
    ## remove that manipulative from targets of other manipulatives:
    for _id, targets in targets_map.items():
        for mid in is_input:
            logger.debug(f"Need to pop: {mid} from {targets}")
            while mid in targets:
                logger.debug(f"popping...{mid} from {_id}")
                targets_map[_id].pop(targets.index(mid))

    # print("PROCESSING OBSERVABLES MANIPULATIVES")
    for mid, x in genome.manipulatives.items():
        if ['observables'] == x['genes']['sources']['value']:
            sources_map[mid] = [x['primitive']]


    logger.debug("sources_map:")
    for k,v in sources_map.items():
        logger.debug(f"{k}:")
        for t in v:
            if t.startswith('action'):
                v.pop(v.index(t))
            else:
                logger.debug(f"    <---{t}")

    logger.debug("targets_map:")
    for k,v in targets_map.items():
        logger.debug(f"{k}:")
        for t in v:
            if t.startswith('action'):
                v.pop(v.index(t))
            else:
                logger.debug(f"    --->{t}")

    logger.debug("as_inputs:")
    for k,v in as_inputs.items():
        logger.debug(f"{k}:")
        for t in v:
            if t.startswith('action'):
                v.pop(v.index(t))
            else:
                logger.debug(f"    --->({t})")

    return sources_map, targets_map, as_inputs

def check_container_conflicts(p_ids, running_container_names):
    docker_client = docker.from_env()
    logger.debug(f'checking container conflicts')
    for _id in p_ids:
        logger.debug(f'  -> {_id}')
        if _id in running_container_names:
            x = input(f"{bg.red}{fg.white} Conflicting containers! {_id} already exists.{bg.rs}{fg.rs} Okay to replace? Y/N: ")
            if x.lower() == 'y':
                c = docker_client.containers.get(_id)
                c.stop()
                c.remove()
            else:
                logger.debug(f"{bg.yellow}{fg.white} Aborting!{fg.rs}{bg.rs}")
                sys.exit()
    return

def find_available_ports():
    """Search all docker containers, looking for free ports to assign
    to GAIuS API, starting at port 8000 (unsecured) and 44300 (secured)

    Returns:
        dict: showing ports found
    """
    docker_client = docker.from_env()
    port_connects={}
    port_offset=0
    port_base = 8000
    port_limit = 8200
    port_list = []
    
    for each in docker_client.containers.list():
        for ports in each.ports.values():
            if ports is None:
                continue
            for port in ports:
                port_list.append(int(port['HostPort']))
    while port_base + port_offset < port_limit:
        if (port_base + port_offset not in port_list) and (44300 + port_offset not in port_list):
            port_connects={'80/tcp': port_base+port_offset, '443/tcp': 44300+port_offset}
            break;
        else:
            port_offset += 1

    if port_base + port_offset >= port_limit:
        logger.debug(f'failed to assign ports: reached port limit')
        sys.exit(1)
    return port_connects

def multi_container_start(genome : Genome, 
            BOTTLE_HOSTNAME=None,
            REGISTRY=None,
            VERSION=None,
            API_KEY=None,
            LOG_LEVEL=None,
            NETWORK=None,
            BIND_PORTS=None,
            user_id = None,
            agent_id = None):

    if user_id is None:
        raise Exception('no user id provided')
    if agent_id is None:
        raise Exception('no agent_id provided')

    docker_client = docker.from_env()
    container_extension = '-' + user_id + '-' +  agent_id

    if not NETWORK:
        NETWORK="g2network"

    net_name = NETWORK + container_extension
    ## Does this network already exist?
    net_name_exists = False
    for n in docker_client.networks.list():
        if n.name == net_name:
            net_name_exists = True
            net = n
    if net_name_exists:
        logger.warning(f"error, {net_name} already exists!")
    else:
        net = docker_client.networks.create(net_name)
        logger.debug(f"{fg.green} Created {net.name} with id: {net.short_id} {fg.rs}")
    logger.debug(net_name)
    logger.debug(net)
    ASSETS['network'][net_name] = net
    
    sources_map, targets_map, as_inputs = parse_genome(genome)

    running_container_names = [c.name for c in docker_client.containers.list()]
    p_ids = list(genome.manipulative_map.keys() ) + list(genome.primitive_map.values()) + ['gaius-api']
    p_ids = [f'{_id}{container_extension}' for _id in p_ids]
    check_container_conflicts(p_ids, running_container_names)

    logger.debug("Starting Manipulative nodes...")
    for _id, m in genome.manipulative_map.items():
        if _id.startswith('action'): ## Needed for old testing topologies.
                continue
        manipulatives_genes = json.dumps(genome.manipulatives[_id])
        if _id in sources_map:
            sources = json.dumps(sources_map[_id])
        else:
            sources = json.dumps([])
        if _id in targets_map:
            targets = json.dumps(targets_map[_id])
        else:
            targets = json.dumps([])

        ASSETS['mp-containers'][_id] = docker_client.containers.run(
            name=_id + container_extension,
            image=f"{REGISTRY}manipulative-processor:{VERSION}",
            network=net_name,
            detach=True,
            init=True,
            hostname=_id,
            restart_policy={"Name": "unless-stopped"},
            environment={'LOG_LEVEL': LOG_LEVEL,
                         'API_KEY': API_KEY,
                         'HOSTNAME': _id,
                         'MANIFEST': manipulatives_genes,
                         'SOURCES': sources,
                         'TARGETS': targets}
            )

    logger.debug("Starting Cognitive Processor nodes...")
    for p, _id in genome.primitive_map.items():
        primitive_genes = json.dumps(genome.primitives[_id])
        if _id in sources_map:
            sources = json.dumps(sources_map[_id])
        else:
            sources = json.dumps([])
        if _id in targets_map:
            targets = json.dumps(targets_map[_id])
        else:
            targets = json.dumps([])
        if _id in as_inputs:
            inputs = json.dumps(as_inputs[_id])
        else:
            inputs = json.dumps([])

        image=f"{REGISTRY}cognitive-processor:{VERSION}"
        logger.debug(f"using cognitive processor image {image}")
        ASSETS['cp-containers'][_id] = docker_client.containers.run(
            name=_id + container_extension,
            image=image,
            network=net_name,
            detach=True,
            init=True,
            hostname=_id,
            restart_policy={"Name": "unless-stopped"},
            environment={'LOG_LEVEL': LOG_LEVEL,
                         'API_KEY': API_KEY,
                         'HOSTNAME': _id,
                         'MANIFEST': primitive_genes,
                         'SOURCES': sources,
                         'AS_INPUTS': inputs,
                         'TARGETS': targets}
            )

    logger.debug("Starting GAIuS-API interface node...")

    port_connects = {}
    if BIND_PORTS:
        port_connects = find_available_ports()
        
    ASSETS['interfaces']['gaius-api' + container_extension] = docker_client.containers.run(name='gaius-api' + container_extension,
            image=f"{REGISTRY}gaius-api:{VERSION}",
            network=net_name,
            detach=True,
            init=True,
            ports=port_connects,
            restart_policy={"Name": "unless-stopped"},
            environment={'LOG_LEVEL': LOG_LEVEL,
                         'API_KEY': API_KEY,
                         'HOSTNAME': BOTTLE_HOSTNAME,
                         'GENOME': json.dumps(genome.topology),
                         'PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION': "cpp"}
            )

    # print summary
    logger.info(f'{fg.green}GAIuS Agent started on docker network {fg.blue}{str(net_name)}{fg.rs}')
    if(BIND_PORTS):
        logger.info(f'{fg.green} Ports {fg.blue}{str(port_connects)}{fg.green} are exposed to the host{fg.rs}')
    else:
        logger.warning(f'{fg.green}Ports were not exposed for this agent. re-run command with {fg.blue}--bind-ports{fg.green} to expose ports{fg.rs}')
    
    return net_name, port_connects

def single_container_start(genome : Genome, 
        BOTTLE_HOSTNAME=None,
        REGISTRY=None,
        VERSION=None,
        API_KEY=None,
        LOG_LEVEL=None,
        NETWORK=None,
        BIND_PORTS=None,
        user_id = None,
        agent_id = None):
    
    if user_id is None:
        raise Exception('no user id provided')
    if agent_id is None:
        raise Exception('no agent_id provided')

    docker_client = docker.from_env()
    container_extension = '-' + user_id + '-' +  agent_id
    
    if not NETWORK:
        NETWORK="g2network"

    net_name = NETWORK + container_extension
    ## Does this network already exist?
    net_name_exists = False
    for n in docker_client.networks.list():
        if n.name == net_name:
            net_name_exists = True
            net = n
    if net_name_exists:
        logger.warning(f"error, {net_name} already exists!")
    else:
        net = docker_client.networks.create(net_name)
        logger.debug(f"{fg.green} Created {net.name} with id: {net.short_id} {fg.rs}")
    logger.debug(net_name)
    logger.debug(net)
    ASSETS['network'][net_name] = net
    
    sources_map, targets_map, as_inputs = parse_genome(genome)

    running_container_names = [genome.topology["agent"]]
    p_ids = list(genome.manipulative_map.keys() ) + list(genome.primitive_map.values()) + ['gaius-api']
    p_ids = [f'{_id}{container_extension}' for _id in p_ids]
    check_container_conflicts(p_ids, running_container_names)

    print("Starting Cognitive Processor nodes...")
    for p, _id in genome.primitive_map.items():
        primitive_genes = json.dumps(genome.primitives[_id])
        if _id in sources_map:
            sources = json.dumps(sources_map[_id])
        else:
            sources = json.dumps([])
        if _id in targets_map:
            targets = json.dumps(targets_map[_id])
        else:
            targets = json.dumps([])
        if _id in as_inputs:
            inputs = json.dumps(as_inputs[_id])
        else:
            inputs = json.dumps([])

    print("Starting GAIuS-API interface node...")
    port_connects={}
    # port_offset = random.randrange(0,500)
    port_offset=0
    port_base = 8000
    port_limit = 8200
    if(BIND_PORTS):

        while port_base + port_offset < port_limit:
            try:
                tmpSocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                # print(f"port offset = {port_offset}")
                avail = tmpSocket.connect_ex(("127.0.0.1", port_base+port_offset))
                # print(f"avail = {avail}")
                if(avail != 0):
                    avail = tmpSocket.connect_ex(("127.0.0.1", 44300+port_offset))
                    # print(f"avail = {avail}")
                    if(avail != 0):
                        print(f"port {port_base+port_offset} not used. using {port_offset} as offset")
                        break
                port_offset = port_offset + 1
                tmpSocket.close()
                if(port_base + port_offset >= port_limit):
                    print(f"ran out of available ports in range. first available port > {port_limit}")
                    sys.exit(-1)
            except ConnectionRefusedError:
                print("Connection refused when searching for ports :(")
                break
            finally:
                tmpSocket.close()
        
        port_connects={'80/tcp': port_base+port_offset, '443/tcp': 44300+port_offset}
        print(port_connects)

    print(f"{genome.primitive_map.keys()}")
    inputs = [val for key, val in genome.primitive_map.items()]
    
    '''
        This section aligns the genes and data according to name, and then
        preprocesses it so that it can be correctly read by the C++ code.
    '''
    # get primitive names
    primitive_names = str([val for key, val in genome.primitive_map.items()])
    primitive_names = primitive_names.replace("[", "")
    primitive_names = primitive_names.replace("]", "")
    primitive_names = primitive_names.replace("'", "")
    primitive_names = primitive_names.replace(" ", "")
    primitive_names = primitive_names.replace(",", ";")

    manip_names = str([key for key in genome.manipulative_map.keys()])
    manip_names = manip_names.replace("[", "")
    manip_names = manip_names.replace("]", "")
    manip_names = manip_names.replace("'", "")
    manip_names = manip_names.replace(" ", "")
    manip_names = manip_names.replace(",", ";")

    print(f"{primitive_names=}")
    print(f"{manip_names=}")

    primitive_genes = str([genome.primitives[key] for key in inputs])

    print(f"{inputs=}")

    primitive_genes = primitive_genes[1:-1]
    # primitive_genes = inputs.replace("'", "")
    primitive_genes = primitive_genes.replace(" ", "")
    primitive_genes = primitive_genes.replace("},", "};")
    primitive_genes = primitive_genes.replace("'", "\"")
    primitive_genes = primitive_genes.replace("False", "false")
    primitive_genes = primitive_genes.replace("True", "true")
    # print(f"{genome.primitive_map.items()=}")
    print(f"{primitive_genes=}")
    print(f"pre {targets_map=}")
    targets = str([targets_map[key] if key in targets_map else [] for key in inputs])
    targets = targets.replace("[[", "[")
    targets = targets.replace("]]", "]")
    targets = targets.replace(" ", "")
    targets = targets.replace("\'", "\"")
    targets = targets.replace(",", ";")
    print(f"post {targets=}")
    print(f"pre {sources_map=}")
    sources = str([sources_map[key] if key in sources_map else [] for key in inputs])
    sources = sources.replace("[[", "[")
    sources = sources.replace("]]", "]")
    sources = sources.replace(" ", "")
    sources = sources.replace("\'", "\"")
    sources = sources.replace("],", "];")
    print(f"post {sources=}")

    print(f"{as_inputs=}")

    inputs = str([[as_inputs[key]] if key in as_inputs else [] for key in inputs])
    inputs = inputs.replace(" ", "")
    inputs = inputs.replace("[[", "[")
    inputs = inputs.replace("]]", "]")
    inputs = inputs.replace("'", "\"")
    inputs = inputs.replace(",", ";")
    inputs = inputs.replace("[[", "[")
    inputs = inputs.replace("]]", "]")

    print(f"final {inputs=}")

    # create manipulative environment variables
    print(f"{genome.manipulatives.keys()=}")
    for val, key in genome.manipulative_map.items():
        print(f"{val=}, {key=}")
        genome.primitive_map[val] = val
        genome.primitives[val] = genome.manipulatives[val]
    
    manipulatives_genes = {}
    manip_ids = []
    manip_sources = {}
    manip_targets = {}
    for _id, m in genome.manipulative_map.items():
        print(f"{_id=}, {m=}")
        manip_ids.append(_id)
        if _id.startswith('action'): ## Needed for old testing topologies.
                continue
        manipulatives_genes[_id] = genome.manipulatives[_id]
        if _id in sources_map:
            manip_sources[_id] = sources_map[_id]
        else:
            manip_sources[_id] = []
        if _id in targets_map:
            manip_targets[_id] = targets_map[_id]
        else:
            manip_targets[_id] = []

    
    manipulatives_genes = [val for val in genome.manipulatives.values()]
    # print(f"{manipulatives_genes=}")

    manipulatives_genes = str(manipulatives_genes)
    print(f"pre processing {manipulatives_genes=}")
    manipulatives_genes = manipulatives_genes[1:-1]
    # primitive_genes = inputs.replace("'", "")
    manipulatives_genes = manipulatives_genes.replace("}, {", "}; {")
    manipulatives_genes = manipulatives_genes.replace(" ", "")
    manipulatives_genes = manipulatives_genes.replace("'", "\"")
    # manipulatives_genes = manipulatives_genes.replace("False", "false")
    # manipulatives_genes = manipulatives_genes.replace("True", "true")
    manip_targets = str([manip_targets[key] for key in manip_ids])
    manip_targets = manip_targets.replace("[[", "[")
    manip_targets = manip_targets.replace("]]", "]")
    manip_targets = manip_targets.replace(" ", "")
    manip_targets = manip_targets.replace("\'", "\"")
    manip_targets = manip_targets.replace(",", ";")
    manip_sources = str([manip_sources[key] for key in manip_ids])
    manip_sources = manip_sources.replace("[[", "[")
    manip_sources = manip_sources.replace("]]", "]")
    manip_sources = manip_sources.replace(" ", "")
    manip_sources = manip_sources.replace("\'", "\"")
    manip_sources = manip_sources.replace("],", "];")
    manip_ids = str(manip_ids)
    manip_ids = manip_ids.replace("[", "")
    manip_ids = manip_ids.replace("]", "")
    manip_ids = manip_ids.replace("'", "")
    manip_ids = manip_ids.replace(" ", "")
    manip_ids = manip_ids.replace(",", ";")
    # print(f"{manipulatives_genes=}")
    print(f"{manip_ids=}")
    print(f"{manip_targets=}")
    print(f"{manip_sources=}")

    str_genome = genome.topology.__str__()

    print(f"{genome.topology.__str__()=}")

    ASSETS['cp-containers'][_id] = docker_client.containers.run(name=genome.topology["agent"],
        image=f"{REGISTRY}connectome:{VERSION}",
        network=net_name,
        detach=True,
        init=True,
        hostname=options.BOTTLE,
        ports=port_connects,
        ipc_mode="shareable",
        privileged=True,
        volumes=["/dev/bus/usb/:/dev/bus/usb/"],
        # publish_all_ports=True,
        restart_policy={"Name": "unless-stopped"},
        environment={'LOG_LEVEL': LOG_LEVEL,
                    'API_KEY': API_KEY,
                    'HOSTNAME': BOTTLE_HOSTNAME,
                    'CONTAINER_NAME': options.BOTTLE,
                    'PRIM_NAMES' : primitive_names,
                    'PRIM_MANIFEST': primitive_genes,
                    'PRIM_SOURCES': sources,
                    'PRIM_AS_INPUTS': inputs,
                    'PRIM_TARGETS': targets,
                    'GENOME': json.dumps(genome.topology),
                    'MANIPNAMES' : manip_names,
                    'MANIPIDS' : manip_ids,
                    'MANIPMANIFEST': manipulatives_genes,
                    'MANIPSOURCES': manip_sources,
                    'MANIPTARGETS': manip_targets}
            )

    # print summary
    logger.info(f'{fg.green}GAIuS Agent started on docker network {fg.blue}{str(net_name)}{fg.rs}')
    if(BIND_PORTS):
        logger.info(f'{fg.green} Ports {fg.blue}{str(port_connects)}{fg.green} are exposed to the host{fg.rs}')
    else:
        logger.warning(f'{fg.green}Ports were not exposed for this agent. re-run command with {fg.blue}--bind-ports{fg.green} to expose ports{fg.rs}')
    
    return net_name, port_connects

def startit(genome_file, 
            BOTTLE_HOSTNAME=None,
            REGISTRY=None,
            VERSION=None,
            API_KEY=None,
            LOG_LEVEL=None,
            NETWORK=None,
            BIND_PORTS=None,
            user_id = None,
            agent_id = None):

    genome = retrieve_genome(genome_file=genome_file)

    print(f"{genome.topology.keys()=}")

    if(genome.topology.get("communication_protocol", "tcp") != "uds"):
        net_name, port_connects = multi_container_start(genome, BOTTLE_HOSTNAME, REGISTRY, VERSION, API_KEY, LOG_LEVEL, NETWORK, BIND_PORTS, user_id, agent_id)
    else:
        net_name, port_connects = single_container_start(genome, BOTTLE_HOSTNAME, REGISTRY, VERSION, API_KEY, LOG_LEVEL, NETWORK, BIND_PORTS, user_id, agent_id)


    return net_name, port_connects

def stopContainer(id):
    try:
        docker_client = docker.from_env()
        logger.info(f'''stopping container {id}''')
        c = docker_client.containers.get(id)
        c.stop()
        c.remove()
        logger.info(f'''container {id} removed{fg.rs}''')
    except:
        logger.debug(f'''{fg.red}container {id} did not exist... continuing{fg.rs}''')

def stopit(genome_file, user_id=None, agent_id=None):

    if user_id is None:
        raise Exception('no user id provided')
    if agent_id is None:
        raise Exception('no agent_id provided')

    container_extension = '-' + user_id + '-' +  agent_id
    docker_client = docker.from_env()

    genome = retrieve_genome(genome_file)

    ids = list(genome.primitive_map.values()) + list(genome.manipulative_map.keys())
    ids.append('gaius-api')

    ids = ['{0}{1}'.format(id, container_extension) for id in ids]
    ids.append(genome.topology["agent"]) # for when single container is run
    multiPool = multiprocessing.Pool()
    multiPool.map(stopContainer, ids)

    try:
        logger.debug("pruning docker system")
        docker_client.containers.prune()
        docker_client.networks.prune()
    except:
        logger.warn("unable to prune docker")
        sys.exit(-1)

    return

def execute_agent(BOTTLE_HOSTNAME   = None,
                  REGISTRY            = 'registry.digitalocean.com/intelligent-artifacts',
                  VERSION             = 'latest',
                  API_KEY             = None,
                  LOG_LEVEL           = 'DEBUG',
                  NETWORK             = None,
                  BIND_PORTS          = True,
                  GENOME              = None,
                  AGENT_ID            = None,
                  USER_ID             = None,
                  KILL                = False
                  ):

    if GENOME is None:
        raise Exception('No genome provided')

    # if not os.path.exists(f"{os.path.expanduser('~')}/.gaius"):
    #     os.mkdir(f"{os.path.expanduser('~')}/.gaius")

    if not API_KEY:
        API_KEY = 'ABCD-1234'

    if not NETWORK:
        NETWORK = 'g2network'

    configuration = {
                    'BOTTLE_HOSTNAME': BOTTLE_HOSTNAME,
                    'REGISTRY': REGISTRY,
                    'VERSION': VERSION,
                    'API_KEY': API_KEY,
                    'LOG_LEVEL': LOG_LEVEL,
                    'NETWORK': NETWORK,
                    'BIND_PORTS': BIND_PORTS,
                    'user_id' : USER_ID,
                    'agent_id' : AGENT_ID,
                    }

    genome_file = GENOME

    if KILL:
        stopit(genome_file, user_id=USER_ID, agent_id=AGENT_ID)
        return -1

    _, ports = startit(genome_file, **configuration)
    
    return ports

def start_agent(json_obj: dict):
    '''Start an agent from a genome passed as JSON

    fields => example:

        genome => json representation of genome
        user_id => test_user
        agent_id => agent1
        kill => False
        version => "latest"

    '''
    
    # required fields
    genome = json_obj['genome']
    user_id = json_obj['user_id']
    agent_id = json_obj['agent_id']
    kill = json_obj['kill']
    network = json_obj['network']
    
    # optional fields
    api_key = json_obj.get('api_key', "ABCD-1234")
    registry = json_obj.get('registry', 'registry.digitalocean.com/intelligent-artifacts/')
    version = json_obj.get('version', 'latest')

    agent_name = 'gaius-api' + "-" + user_id + "-" + agent_id

    ports = execute_agent(REGISTRY=registry,
                          BIND_PORTS=True,
                          AGENT_ID=agent_id,
                          USER_ID=user_id,
                          KILL=kill,
                          API_KEY=api_key,
                          GENOME=json.dumps(genome),
                          VERSION=version,
                          NETWORK=network)

    if kill:
        return -1
    
    logger.debug(f'started agent with ports {ports}')

    agent_info =  {"name": "",
                   "domain": agent_name,
                   "api_key":api_key,
                   "secure": False}

    return agent_info, ports

def options():
    global options
    parser = OptionParser(version=ia.__version__)
    parser.add_option("-g", "--genome", dest="filename",
                        help="Genome file location", metavar="FILE")

    parser.add_option("--api-key", dest="API_KEY", default='ABCD-1234',
                        help="""Secret API key.
                        Default = 'ABCD-1234'
                                """)

    parser.add_option("--log-level", dest="LOG_LEVEL", default='DEBUG',
                        help="""Log level.
                        Default = DEBUG
                        Options:
                            CRITICAL
                            ERROR
                            WARNING
                            INFO
                            DEBUG
                                """)

    parser.add_option("-n", "--network", dest="NETWORK",
                        help="""Network in which all nodes will communicate.
                        Default = 'g2network'
                                """)
    parser.add_option("--bind-ports", action="store_true", dest="BIND_PORTS", default=False,
                        help="""To enable multiple agents running on single system, port are not bound by default.
                        Specify this option to explicitly bind agent to ports (required to run tests on agent).
                            HTTP port: 8000 + agent_id
                            HTTPS port: 44300 + agent_id
                            """)

    parser.add_option("-b", "--gaius-api-name", dest="BOTTLE", default='gaius-api',
                        help="""Bottle name
                        Default = 'BOTTLE'
                                """)

    parser.add_option("-H", "--hostname", dest="BOTTLE_HOSTNAME", default='localhost',
                        help="""Bottle name
                        Default = 'BOTTLE'
                                """)

    parser.add_option("-r", "--registry", dest="REGISTRY", default='registry.digitalocean.com/intelligent-artifacts/',
                        help="""Container registry
                        Default = 'registry.digitalocean.com/intelligent-artifacts/'

                        Set to '' if you want to use only local containers.
                                """)

    parser.add_option("--container-version", dest="VERSION", default='latest',
                        help="""Version of container images. Same used for all.
                        Default = 'latest'
                        """)

    parser.add_option("-d", "--debug",
                        action="store_true", dest="debug", default=False,
                        help="print debug information")

    parser.add_option("--kill",
                        action="store_true", dest="kill", default=False,
                        help="Kill all agent's containers.")

    parser.add_option("--agent-id",
                        dest="agent_id", default="1",
                        help="""A string that uniquely identifies the agent and network. Appended to container and network names
                             Default=1""")
    
    parser.add_option("--user-id",
                        dest="user_id", default=os.environ.get("USER"),
                        help="""A string that uniquely identifies the agent and network. Appended to container and network names
                             Default=os.environ.get("USER")""")
    
    (options, args) = parser.parse_args()

    if not options.filename:
        logger.warn("Need a genome file location! Pass with -g or --genome arg.")
        sys.exit(1)

    if len(options.REGISTRY) > 0 and not options.REGISTRY.endswith('/'):
        options.REGISTRY += '/'

if __name__ == '__main__':
    # if not os.path.exists(f"{os.path.expanduser('~')}/.gaius"):
    #     os.mkdir(f"{os.path.expanduser('~')}/.gaius")
    options()

    DEBUG = options.debug
    if DEBUG:
        logger.setLevel(level=logging.DEBUG)

    ## Default values provided by OptionParser
    BOTTLE_HOSTNAME = options.BOTTLE_HOSTNAME
    REGISTRY = options.REGISTRY
    VERSION = options.VERSION
    API_KEY = options.API_KEY
    LOG_LEVEL = options.LOG_LEVEL
    if(options.NETWORK):
        NETWORK = options.NETWORK
    else:
        NETWORK = "g2network"

    BIND_PORTS = options.BIND_PORTS

    USER_ID = options.user_id
    AGENT_ID = options.agent_id
    configuration = {
                    'BOTTLE_HOSTNAME': BOTTLE_HOSTNAME,
                    'REGISTRY': REGISTRY,
                    'VERSION': VERSION,
                    'API_KEY': API_KEY,
                    'LOG_LEVEL': LOG_LEVEL,
                    'NETWORK': NETWORK,
                    'BIND_PORTS': BIND_PORTS,
                    'user_id' : USER_ID,
                    'agent_id' : AGENT_ID,
                    }

    genome_file = options.filename
    if options.kill:
        stopit(genome_file, user_id=USER_ID, agent_id=AGENT_ID)
        sys.exit(0)

    network = startit(genome_file, **configuration)